/****************************************************************************
 * This is an XDR data description of C ASTs. The structure defined here is *
 * very similar to that of CIL, and instances are intended to be generated  *
 * from CIL using the Writexdr module (command line option --dowritexdr).   *
 *                                                                          *
 * A complete XDR serialization also requires a separate file containing a  *
 * table of interned strings (which is also generated by Writexdr). See     *
 * string_table.x for a description of that file.                           *
 *                                                                          *
 * You should be familiar with the following before reading this:           *
 *    XDR: http://tools.ietf.org/html/rfc4506                               *
 *    CIL: http://hal.cs.berkeley.edu/cil/                                  *
 ***************************************************************************/

/* Whenever a string would appear in the AST, its ID is encoded instead. To
 * find the string, just look up the ID in the string table. */
typedef unsigned int string_id;

/****************************************************************************
 * FILES AND GLOBALS                                                        *
 ***************************************************************************/

/* A file is the root of the AST. */
struct file {
   string_id         file_name;
   global            file_globals<>;
};

/* A typedef. Occurs only at global level. */
struct type_defn {
   string_id         type_defn_name;
   bool              type_defn_used;
   c_type            type_defn_type;
   location          type_defn_loc;
};

/* Because structs and unions are so similar, they are both treated as
 * "composites", and this flag is stored to indicate what the composite is
 * representing. */
enum comp_kind {
   COMPOSITE_STRUCT = 0,
   COMPOSITE_UNION = 1
};

/* A field declaration. These appear only in composite definitions. */
struct field_decl {
   string_id         field_decl_name;
   int *             field_decl_bit_width;
   c_type            field_decl_type;
   attribute         field_decl_attrs<>;
   location          field_decl_loc;
};

/* A composite (struct/union) definition. Occurs only at global level. */
struct comp_defn {
   string_id         comp_defn_name;
   comp_kind         comp_defn_kind;
   int               comp_defn_key;
   bool              comp_defn_used;
   field_decl        comp_defn_fields<>;
   attribute         comp_defn_attrs<>;
   location          comp_defn_loc;
};

/* A composite (struct/union) forward declaration. Occurs only at global
 * level. */
struct comp_decl {
   string_id         comp_decl_name;
   comp_kind         comp_decl_kind;
   int               comp_decl_key;
   bool              comp_decl_defined;
   bool              comp_decl_used;
   attribute         comp_decl_attrs<>;
   location          comp_decl_loc;
};

/* An enum item. Appears only in enum definitions. */
struct enum_item {
   string_id         enum_item_name;
   expr              enum_item_value;
   location          enum_item_loc;
};

/* An enum definition. Occurs only at global level. */
struct enum_defn {
   string_id         enum_defn_name;
   bool              enum_defn_used;
   enum_item         enum_defn_items<>;
   attribute         enum_defn_attrs<>;
   location          enum_defn_loc;
};

/* An enum declaration. Occurs only at global level. */
struct enum_decl {
   string_id         enum_decl_name;
   bool              enum_decl_used;
   attribute         enum_decl_attrs<>;
   location          enum_decl_loc;
};

/* Storage class information. Appears in variable and function definitions
 * and declarations. */
enum storage_class {
   NO_STORAGE = 0,
   STATIC = 1,
   REGISTER = 2,
   EXTERN = 3
};

/* A variable definition, possibly with an initializer. Local variable
 * definitions with initalizers are transformed into declarations and
 * subsequent assignments, so this occurs only at global level. */
struct var_defn {
   string_id         var_defn_name;
   int               var_defn_id;
   bool              var_defn_address_taken;
   bool              var_defn_used;
   c_type            var_defn_type;
   storage_class     var_defn_storage;
   attribute         var_defn_attrs<>;
   location          var_defn_loc;
   init *            var_defn_initializer;
};

/* A variable declaration. Used for global, function parameter, and local
 * variable declarations. */
struct var_decl {
   string_id         var_decl_name;
   int               var_decl_id;
   bool              var_decl_global;
   bool              var_decl_address_taken;
   bool              var_decl_used;
   c_type            var_decl_type;
   storage_class     var_decl_storage;
   attribute         var_decl_attrs<>;
   location          var_decl_loc;
};

/* A block of statements. Really just a statement sequence, but it can also
 * have its own attributes. Appears in function definitions and in many
 * instructions. */
struct block {
   attribute         block_attrs<>;
   stmt              block_stmts<>;
};

/* A function definition. Occurs only at global level (GCC nested functions
 * are not supported). */
struct func_defn {
   string_id         func_defn_name;
   int               func_defn_id;
   bool              func_defn_inline;
   bool              func_defn_address_taken;
   bool              func_defn_used;
   c_type            func_defn_type;
   storage_class     func_defn_storage;
   var_decl          func_defn_params<>;
   var_decl          func_defn_locals<>;
   block             func_defn_body;
   attribute         func_defn_attrs<>;
   location          func_defn_loc;
};

/* A function declaration (a prototype). Occurs only at global level. */
struct func_decl {
   string_id         func_decl_name;
   int               func_decl_id;
   bool              func_decl_inline;
   bool              func_decl_address_taken;
   bool              func_decl_used;
   c_type            func_decl_type;
   storage_class     func_decl_storage;
   attribute         func_decl_attrs<>;
   location          func_decl_loc;
};

/* A global assembly statement. Much less complicated than local inline asm;
 * this can contain only a template. */
struct glob_asm {
   string_id         glob_asm_template;
   location          glob_asm_loc;
};

/* A pragma at global level. */
struct pragma {
   attribute         pragma_content;
   location          pragma_loc;
};

/* Discriminator for the union representation of globals. */
enum global_kind {
   GLOBAL_TYPE_DEFN = 0,
   GLOBAL_COMPOSITE_TAG_DEFN = 1,
   GLOBAL_COMPOSITE_TAG_DECL = 2,
   GLOBAL_ENUM_TAG_DEFN = 3,
   GLOBAL_ENUM_TAG_DECL = 4,
   GLOBAL_VARIABLE_DEFN = 5,
   GLOBAL_VARIABLE_DECL = 6,
   GLOBAL_FUNCTION_DEFN = 7,
   GLOBAL_FUNCTION_DECL = 8,
   GLOBAL_ASM_STATEMENT = 9,
   GLOBAL_PRAGMA = 10,
   GLOBAL_TEXT = 11
};

/* A global. A file is basically just a named list of these. */
union global switch (global_kind k) {
   case GLOBAL_TYPE_DEFN:
      type_defn      global_type_defn;
   case GLOBAL_COMPOSITE_TAG_DEFN:
      comp_defn      global_composite_tag_defn;
   case GLOBAL_COMPOSITE_TAG_DECL:
      comp_decl      global_composite_tag_decl;
   case GLOBAL_ENUM_TAG_DEFN:
      enum_defn      global_enum_tag_defn;
   case GLOBAL_ENUM_TAG_DECL:
      enum_decl      global_enum_tag_decl;
   case GLOBAL_VARIABLE_DEFN:
      var_defn       global_variable_defn;
   case GLOBAL_VARIABLE_DECL:
      var_decl       global_variable_decl;
   case GLOBAL_FUNCTION_DEFN:
      func_defn      global_function_defn;
   case GLOBAL_FUNCTION_DECL:
      func_decl      global_function_decl;
   case GLOBAL_ASM_STATEMENT:
      glob_asm       global_asm_statement;
   case GLOBAL_PRAGMA:
      pragma         global_pragma;
   case GLOBAL_TEXT:
      string_id      global_text;
};

/****************************************************************************
 * TYPES                                                                    *
 ***************************************************************************/

/* The various kinds of integers in C. Appears in types and constants. */
enum int_kind {
   CHAR = 0,
   SIGNED_CHAR = 1,
   UNSIGNED_CHAR = 2,
   INT = 3,
   UNSIGNED_INT = 4,
   SHORT = 5,
   UNSIGNED_SHORT = 6,
   LONG = 7,
   UNSIGNED_LONG = 8,
   LONG_LONG = 9,
   UNSIGNED_LONG_LONG = 10
};

/* The various kinds of floating-point numbers in C. Appears in types and
 * constants. */
enum float_kind {
   FLOAT = 0,
   DOUBLE = 1,
   LONG_DOUBLE = 2
};

/* Everything below this comment and above the definition of c_type are only
 * "helper" structures for c_type. They appear nowhere else. */

/* Integral types. The specific kind of integer (short, int, etc.) is stored
 * in the int_kind. */
struct c_int {
   int_kind          c_int_kind;
   attribute         c_int_attrs<>;
};

/* Floating-point types. The level of precision (float, double, long double)
 * is stored in the float_kind. */
struct c_float {
   float_kind        c_float_kind;
   attribute         c_float_attrs<>;
};

/* Pointer types. */
struct c_ptr {
   c_type            c_ptr_pointed_to;
   attribute         c_ptr_attrs<>;
};

/* Array types. */
struct c_array {
   c_type            c_array_members;
   int *             c_array_length;
   attribute         c_array_attrs<>;
};

/* Each parameter in a function type can have its own attributes. */
struct parameter {
   c_type            parameter_type;
   attribute         parameter_attrs<>;
};

/* The list of parameters in a function can take 3 forms. */
enum parameter_list_kind {
   NO_PARAMETERS = 0,                  /* foo(void); */
   SOME_PARAMETERS = 1,                /* foo(some params); */
   PARAMETERS_UNKNOWN = 2              /* prototype of foo not seen yet */
};

/* The variation in parameter lists is represented with a union. */
union parameter_list switch (parameter_list_kind k) {
   case NO_PARAMETERS:
      void;
   case SOME_PARAMETERS:
      parameter parameters<>;
   case PARAMETERS_UNKNOWN:
      void;
};

/* Function types. */
struct c_fun {
   bool              c_fun_varargs;
   c_type            c_fun_result;
   parameter_list    c_fun_formals;
   attribute         c_fun_attrs<>;
};

/* Named types (i.e., those defined by a typedef). */
struct c_named {
   string_id         c_named_name;
   attribute         c_named_attrs<>;
};

/* Composite types (i.e., structs or unions). */
struct c_comp {
   string_id         c_comp_name;
   attribute         c_comp_attrs<>;
};

/* Enumeration types. */
struct c_enum {
   string_id         c_enum_name;
   attribute         c_enum_attrs<>;
};

/* Discriminator for the union representation of C types. */
enum c_type_kind {
   TYPE_VOID = 0,
   TYPE_INT = 1,
   TYPE_FLOAT = 2,
   TYPE_POINTER = 3,
   TYPE_ARRAY = 4,
   TYPE_FUNCTION = 5,
   TYPE_NAMED = 6,
   TYPE_COMPOSITE = 7,
   TYPE_ENUMERATED = 8,
   TYPE_GCC_BUILTIN_VA_LIST = 9
};

/* A C type. Appears in many places, wherever type information is stored. */
union c_type switch (c_type_kind k) {
   case TYPE_VOID:
      attribute      attrs<>;
   case TYPE_INT:
      c_int          type_int;
   case TYPE_FLOAT:
      c_float        type_float;
   case TYPE_POINTER:
      c_ptr          type_pointer;
   case TYPE_ARRAY:
      c_array        type_array;
   case TYPE_FUNCTION:
      c_fun          type_function;
   case TYPE_NAMED:
      c_named        type_named;
   case TYPE_COMPOSITE:
      c_comp         type_composite;
   case TYPE_ENUMERATED:
      c_enum         type_enumerated;
   case TYPE_GCC_BUILTIN_VA_LIST:
      attribute      attrs<>;
};

/****************************************************************************
 * ATTRIBUTES                                                               *
 ***************************************************************************/

/* An attribute is just a name with some parameters. The rest of this section
 * is devoted to describing the parameters. Attributes can be applied all
 * over the place and appear almost everywhere. */
struct attribute {
   string_id         attribute_name;
   attr_param        attribute_params<>;
};

/* Everything below this comment and above the definition of attr_param are
 * only "helper" structures for attr_param. They appear nowhere else. */

/* Constructed attributes. */
struct attr_cons {
   string_id         attr_cons_name;
   attr_param        attr_cons_params<>;
};

/* A unary operator applied to an attribute. The actual operator is stored
 * in unary_op_kind. */
struct attr_unary {
   unary_op_kind     attr_unary_operator;
   attr_param        attr_unary_param;
};

/* A binary operator applied to two attributes. The actual operator is stored
 * in binary_op_kind. */
struct attr_binary {
   binary_op_kind    attr_binary_operator;
   attr_param        attr_binary_first_param;
   attr_param        attr_binary_second_param;
};

/* A "dot" applied to an attribute, as in "foo.bar". */
struct attr_dot {
   attr_param        attr_dot_param;
   string_id         attr_dot_name;
};

/* An index operation applied to an attribute, as in "foo[bar]". */
struct attr_index {
   attr_param        attr_index_base;
   attr_param        attr_index_offset;
};

/* The ternary operator ?: applied to three attributes. */
struct attr_ternary {
   attr_param        attr_ternary_guard;
   attr_param        attr_ternary_true_branch;
   attr_param        attr_ternary_false_branch;
};

/* Discriminator for the union representation of attribute parameters. */
enum attr_param_kind {
   PARAM_INT = 0,
   PARAM_STRING = 1,
   PARAM_CONSTRUCTED = 2,
   PARAM_SIZEOF_TYPE = 3,
   PARAM_SIZEOF_PARAM = 4,
   PARAM_ALIGNOF_TYPE = 5,
   PARAM_ALIGNOF_PARAM = 6,
   PARAM_UNARY_OP = 7,
   PARAM_BINARY_OP = 8,
   PARAM_DOT = 9,
   PARAM_STAR = 10,
   PARAM_ADDROF = 11,
   PARAM_INDEX = 12,
   PARAM_TERNARY_OP = 13
};

/* Attribute parameters. */
union attr_param switch (attr_param_kind k) {
   case PARAM_INT:
      int            param_int;
   case PARAM_STRING:
      string_id      param_string;
   case PARAM_CONSTRUCTED:
      attr_cons      param_constructed;
   case PARAM_SIZEOF_TYPE:
      c_type         param_sizeof_type;
   case PARAM_SIZEOF_PARAM:
      attr_param     param_sizeof_param;
   case PARAM_ALIGNOF_TYPE:
      c_type         param_alignof_type;
   case PARAM_ALIGNOF_PARAM:
      attr_param     param_alignof_param;
   case PARAM_UNARY_OP:
      attr_unary     param_unary_op;
   case PARAM_BINARY_OP:
      attr_binary    param_binary_op;
   case PARAM_DOT:
      attr_dot       param_dot;
   case PARAM_STAR:
      attr_param     param_star;
   case PARAM_ADDROF:
      attr_param     param_addrof;
   case PARAM_INDEX:
      attr_index     param_index;
   case PARAM_TERNARY_OP:
      attr_ternary   param_ternary_op;
};

/****************************************************************************
 * EXPRESSIONS (SIDE-EFFECT FREE)                                           *
 ***************************************************************************/

/* Everything below this comment and above the definition of expr are only
 * "helper" structures for expr. They appear nowhere else. */

/* A unary operator applied to an expression. The actual operator is stored
 * in unary_op_kind. */
struct expr_unary {
   unary_op_kind     expr_unary_operator;
   expr              expr_unary_operand;
   c_type            expr_unary_type;
};

/* A binary operator applied to two expressions. The actual operator is
 * stored in binary_op_kind. */
struct expr_binary {
   binary_op_kind    expr_binary_operator;
   expr              expr_binary_first_operand;
   expr              expr_binary_second_operand;
   c_type            expr_binary_type;
};

/* A cast of an expression to a different type. */
struct expr_cast {
   c_type            expr_cast_type;
   expr              expr_cast_expr;
};

/* Discriminator for the union representation of expressions. */
enum expr_kind {
   EXPR_CONST = 0,
   EXPR_LVALUE = 1,
   EXPR_SIZEOF_TYPE = 2,
   EXPR_SIZEOF_EXPR = 3,
   EXPR_SIZEOF_STRING_LITERAL = 4,
   EXPR_ALIGNOF_TYPE = 5,
   EXPR_ALIGNOF_EXPR = 6,
   EXPR_UNARY_OP = 7,
   EXPR_BINARY_OP = 8,
   EXPR_TYPE_CAST = 9,
   EXPR_ADDROF = 10,
   EXPR_STARTOF = 11
};

/* Expressions (side-effect free C statements that return a value). They
 * occur in many places. */
union expr switch (expr_kind k) {
   case EXPR_CONST:
      constant       expr_const;
   case EXPR_LVALUE:
      lval           expr_lvalue;
   case EXPR_SIZEOF_TYPE:
      c_type         expr_sizeof_type;
   case EXPR_SIZEOF_EXPR:
      expr           expr_sizeof_expr;
   case EXPR_SIZEOF_STRING_LITERAL:
      string_id      expr_sizeof_string_literal;
   case EXPR_ALIGNOF_TYPE:
      c_type         expr_alignof_type;
   case EXPR_ALIGNOF_EXPR:
      expr           expr_alignof_expr;
   case EXPR_UNARY_OP:
      expr_unary     expr_unary_op;
   case EXPR_BINARY_OP:
      expr_binary    expr_binary_op;
   case EXPR_TYPE_CAST:
      expr_cast      expr_type_cast;
   case EXPR_ADDROF:
      lval           expr_addrof;
   case EXPR_STARTOF:
      lval           expr_startof;
};

/****************************************************************************
 * CONSTANTS                                                                *
 ***************************************************************************/

/* Everything below this comment and above the definition of constant are
 * only "helper" structures for constant. They appear nowhere else. */

/* An integer constant along with its source text representation. */
struct int_num {
   hyper             int_num_value;
   int_kind          int_num_kind;
   string_id *       int_num_text_rep;
};

/* A floating-point constant along with its source text representation. */
struct float_num {
   double            float_num_value;
   float_kind        float_num_kind;
   string_id *       float_num_text_rep;
};

/* A constant named in an enum. */
struct enum_val {
   string_id         enum_val_name;
   expr              enum_val_value;
   string_id         enum_val_host_enum_name;
};

enum constant_kind {
   CONST_INTEGER = 0,
   CONST_STRING = 1,
   CONST_WIDE_STRING = 2,
   CONST_CHAR = 3,
   CONST_FLOATING_POINT = 4,
   CONST_ENUMERATION = 5
};

union constant switch (constant_kind k) {
   case CONST_INTEGER:
      int_num        const_integer;
   case CONST_STRING:
      string_id      const_string;
   case CONST_WIDE_STRING:
      hyper          const_wide_string<>;
   case CONST_CHAR:
      int            const_char;
   case CONST_FLOATING_POINT:
      float_num      const_floating_point;
   case CONST_ENUMERATION:
      enum_val       const_enumeration;
};

/****************************************************************************
 * OPERATORS                                                                *
 ***************************************************************************/

/* The unary operators. */
enum unary_op_kind {
   UNOP_NEGATION = 0,
   UNOP_BITWISE_COMPLEMENT = 1,
   UNOP_LOGICAL_NOT = 2
};

/* The binary operators. */
enum binary_op_kind {
   BINOP_ARITHMETIC_PLUS = 0,
   BINOP_POINTER_PLUS_INTEGER = 1,
   BINOP_POINTER_PLUS_INTEGER_FROM_ARRAY = 2,
   BINOP_ARITHMETIC_MINUS = 3,
   BINOP_POINTER_MINUS_INTEGER = 4,
   BINOP_POINTER_MINUS_POINTER = 5,
   BINOP_MULTIPLICATION = 6,
   BINOP_DIVISION = 7,
   BINOP_MODULUS = 8,
   BINOP_SHIFT_LEFT = 9,
   BINOP_SHIFT_RIGHT = 10,
   BINOP_LESS_THAN = 11,
   BINOP_GREATER_THAN = 12,
   BINOP_LESS_OR_EQUAL = 13,
   BINOP_GREATER_OR_EQUAL = 14,
   BINOP_EQUAL = 15,
   BINOP_NOT_EQUAL = 16,
   BINOP_BITWISE_AND = 17,
   BINOP_BITWISE_XOR = 18,
   BINOP_BITWISE_OR = 19,
   BINOP_LOGICAL_AND = 20,
   BINOP_LOGICAL_OR = 21
};

/****************************************************************************
 * LVALUES                                                                  *
 ***************************************************************************/

/* An lvalue is a base object and an optional memory offset from it. */
struct lval {
   host              lval_host;
   offset            lval_offset;
};

/* Discriminator for the union representation of hosts. */
enum host_kind {
   HOST_VARIABLE = 0,
   HOST_OBJECT_AT_EFFECTIVE_ADDRESS = 1
};

/* A host is either an object directly referenced by a variable, or an object
 * indirectly referenced by a pointer expression. Hosts appear only in
 * lvalues. */
union host switch (host_kind k) {
   case HOST_VARIABLE:
      int            host_var_id;
   case HOST_OBJECT_AT_EFFECTIVE_ADDRESS:
      expr           host_object_at_effective_address;
};

/* Use of the . operator to access a field of a composite. This occurs only
 * in lvalue offsets. */
struct field_acc {
   string_id         field_acc_field_name;
   string_id         field_acc_host_comp_name;
};

/* A field access offset, along with the next one in the list. */
struct ofst_field {
   field_acc         ofst_field_acc;
   offset            ofst_field_next;
};

/* An array index offset, along with the next one in the list. */
struct ofst_array {
   expr              ofst_array_index;
   offset            ofst_array_next;
};

/* Discriminator for the union representation of offsets. */
enum offset_kind {
   OFFSET_NONE = 0,
   OFFSET_FIELD_ACCESS = 1,
   OFFSET_ARRAY_INDEX = 2
};

/* An lvalue offset, represented like a linked list. Either there is no
 * offset (terminates the list), or there is a field access of a composite
 * or an index into an array. Each of these changes the lvalue to identify a
 * memory location other than the base object, and allows for further offsets
 * (there is often a string of them, such as "foo.bar.baz[qux]"). Offsets
 * appear in lvalues, but are also used in initializers. */
union offset switch (offset_kind k) {
   case OFFSET_NONE:
      void;
   case OFFSET_FIELD_ACCESS:
      ofst_field     offset_field_access;
   case OFFSET_ARRAY_INDEX:
      ofst_array     offset_array_index;
};

/****************************************************************************
 * INITIALIZERS FOR GLOBAL VARIABLES                                        *
 ***************************************************************************/

/* An initializer with the field or index being initialized represented as an
 * lvalue offset. The offset is always an ofst_field or an ofst_array, and
 * its "next offset" field is always void. Appears only in comp_init. */
struct init_with_ofst {
   offset            init_with_ofst_offset;
   init              init_with_ofst_init;
};

/* A compound initializer with a list of fields or indecies to be
 * initialized. Appears only in init. */
struct comp_init {
   c_type            comp_init_type;
   init_with_ofst    comp_init_inits<>;
};

/* Discriminator for union representation of initializers. */
enum init_kind {
   INIT_SINGLE = 0,
   INIT_COMPOUND = 1
};

/* Initializers for global variables. (Initializers for local variables are
 * turned into assignments.) These can either be single initializers for
 * scalar variables, or compound initializers for struct, union, and array
 * variables. */
union init switch (init_kind k) {
   case INIT_SINGLE:
      expr           init_single;
   case INIT_COMPOUND:
      comp_init      init_compound;
};

/****************************************************************************
 * LABELS AND STATEMENTS                                                    *
 ***************************************************************************/

/* A real label. Might be in the original source, or might be generated by
 * CIL. */
struct real_lbl {
   string_id         real_lbl_text;
   location          real_lbl_loc;
   bool              real_lbl_in_orig_source;
};

/* A label that is really a case statement in a switch. */
struct case_lbl {
   expr              case_lbl_expr;
   location          case_lbl_loc;
};

/* Discriminator for the union representation of labels. */
enum label_kind {
   LABEL_REAL_LABEL = 0,
   LABEL_CASE_STATEMENT = 1,
   LABEL_DEFAULT_STATEMENT = 2
};

/* A label. A statement may have one or more of these. */
union label switch (label_kind k) {
   case LABEL_REAL_LABEL:
      real_lbl       label_real_label;
   case LABEL_CASE_STATEMENT:
      case_lbl       label_case_statement;
   case LABEL_DEFAULT_STATEMENT:
      location       label_default_statement;
};

/* A statement. These make up the control-flow graph. A statement can have
 * zero, one, or many labels, and if the CFG has been generated, it has a
 * >= 0 ID and lists of its successors and predecessors. If not, the ID is
 * -1, and the lists are empty. The actual kind and content of the statement
 * is stored in real_stmt. */
struct stmt {
   label             stmt_labels<>;
   real_stmt         stmt_content;
   int               stmt_cfg_id;
   stmt              stmt_cfg_successors<>;
   stmt              stmt_cfg_predecessors<>;
};

/* The rest of this section is devoted to real_stmt and its "helper" types,
 * which appear nowhere else. */

/* A return statement, optionally with a value. */
struct return_stmt {
   expr *            return_stmt_value;
   location          return_stmt_loc;
};

/* A goto statement. Since storing the actual target statement here could
 * cause an infinite loop in serialization, the first label in its labels
 * list is stored instead. You can look up the statement given that label,
 * since it must be unique within the file. */
struct goto_stmt {
   label             goto_stmt_target_label;
   location          goto_stmt_loc;
};

/* An if statement. */
struct if_stmt {
   expr              if_stmt_guard;
   block             if_stmt_then_branch;
   block             if_stmt_else_branch;
   location          if_stmt_loc;
};

/* A switch statement. The cases are given both in the provided block and in
 * the stmt list. */
struct switch_stmt {
   expr              switch_stmt_guard;
   block             switch_stmt_block;
   stmt              switch_stmt_cases<>;
   location          switch_stmt_loc;
};

/* An infinite loop statement (i.e., while (1)). CIL transforms all loops
 * into this loop and uses a break statement to implement the termination
 * test. If the CFG has been prepared, the statements containing the
 * continue and break labels for the loop are given. */
struct loop_stmt {
   block             loop_stmt_body;
   stmt *            loop_stmt_cfg_continue;
   stmt *            loop_stmt_cfg_break;
   location          loop_stmt_loc;
};

/* An MSVC try-finally sequence. */
struct try_finally {
   block             try_finally_try;
   block             try_finally_finally;
   location          try_finally_loc;
};

/* The filter for an MSVC try-except sequence. There is a sequence of
 * (presumably side-effect-ful) instructions whose values are discarded,
 * and then the given expression is used as the real value for the filter.
 * Don't ask me, man. */
struct filter_seq {
   instr             filter_seq_discarded_values<>;
   expr              filter_seq_real_value;
};

/* An MSVC try-except sequence. */
struct try_except {
   block             try_except_try;
   filter_seq        try_except_filter;
   block             try_except_except;
   location          try_except_loc;
};

/* Discriminator for the union type of statements. */
enum real_stmt_kind {
   STMT_INSTRUCTIONS = 0,
   STMT_RETURN = 1,
   STMT_GOTO = 2,
   STMT_BREAK = 3,
   STMT_CONTINUE = 4,
   STMT_IF = 5,
   STMT_SWITCH = 6,
   STMT_INFINITE_LOOP = 7,
   STMT_BLOCK = 8,
   STMT_MSVC_TRY_FINALLY = 9,
   STMT_MSVC_TRY_EXCEPT = 10
};

/* A statement. This is either a sequence of instructions that do not modify
 * control flow (so control implicitly falls through), or statement such as
 * if/switch that does modify control flow. */
union real_stmt switch (real_stmt_kind k) {
   case STMT_INSTRUCTIONS:
      instr          stmt_instructions<>;
   case STMT_RETURN:
      return_stmt    stmt_return;
   case STMT_GOTO:
      goto_stmt      stmt_goto;
   case STMT_BREAK:
      location       stmt_break;
   case STMT_CONTINUE:
      location       stmt_continue;
   case STMT_IF:
      if_stmt        stmt_if;
   case STMT_SWITCH:
      switch_stmt    stmt_switch;
   case STMT_INFINITE_LOOP:
      loop_stmt      stmt_infinite_loop;
   case STMT_BLOCK:
      block          stmt_block;
   case STMT_MSVC_TRY_FINALLY:
      try_finally    stmt_msvc_try_finally;
   case STMT_MSVC_TRY_EXCEPT:
      try_except     stmt_msvc_try_except;
};

/****************************************************************************
 * INSTRUCTIONS                                                             *
 ***************************************************************************/

/* An assignment statement. Appears only in instr. */
struct assignment {
   lval              assignment_lvalue;
   expr              assignment_contents;
   location          assignment_loc;
};

/* A function invocation, possibly with an lvalue to put the result in.
 * Appears only in instr. */
struct fun_call {
   lval *            fun_call_result_lvalue;
   expr              fun_call_name;
   expr              fun_call_args<>;
   location          fun_call_loc;
};

/* Inline assembly output operands. */
struct asm_out {
   string_id *       asm_out_name;
   string_id         asm_out_constraint;
   lval              asm_out_destination;
};

/* Inline assembly input operands. */
struct asm_in {
   string_id *       asm_in_name;
   string_id         asm_in_constraint;
   expr              asm_in_loaded_input;
};

/* An inline assembly statement, following the GCC specification. */
struct inline_asm {
   bool              inline_asm_volatile;
   string_id         inline_asm_templates<>;
   asm_out           inline_asm_output_operands<>;
   asm_in            inline_asm_input_operands<>;
   string_id         inline_asm_clobbered_registers<>;
   location          inline_asm_loc;
};

/* Discriminator for the union representation of instructions. */
enum instr_kind {
   INSTR_ASSIGNMENT = 0,
   INSTR_FUNCTION_CALL = 1,
   INSTR_INLINE_ASM = 2
};

/* An instruction (a statement with no intraprocedural control flow). It is
 * either an assignment, a function call, or inline assembly. */
union instr switch (instr_kind k) {
   case INSTR_ASSIGNMENT:
      assignment     instr_assignment;
   case INSTR_FUNCTION_CALL:
      fun_call       instr_function_call;
   case INSTR_INLINE_ASM:
      inline_asm     instr_inline_asm;
};

/****************************************************************************
 * MISCELLANEOUS                                                            *
 ***************************************************************************/

/* The location of a global, label, statement, or instruction in the original
 * source file. Would be printed in code as a #line directive. */
struct location {
   int line;
   string file<>;
   int byte;
};
