open Printf
open Rtypes

let ic = ref stdin

let comps = Hashtbl.create 1024
let fields = Hashtbl.create 1024
let enums = Hashtbl.create 1024
let types = Hashtbl.create 1024
let vars = Hashtbl.create 1024
let stmts = Hashtbl.create 1024

(* An exact size for the string table will be known to read_string_table, so
 * defer its creation until then. *)
let strings = ref None

(****************************************************************************
 * READERS FOR THE NECESSARY XDR PRIMITIVE TYPES                            *
 ***************************************************************************)

let read_int () =
  let buf = String.create 4 in
  really_input !ic buf 0 4;
  int_of_int4 (read_int4 buf 0)

let read_uint () =
  let buf = String.create 4 in
  really_input !ic buf 0 4;
  int_of_uint4 (read_uint4 buf 0)

let read_bool () = match read_int () with
  | 0 -> false
  | 1 -> true
  | _ -> assert false

let read_hyper () =
  let buf = String.create 8 in
  really_input !ic buf 0 8;
  int64_of_int8 (read_int8 buf 0)

let read_double () =
  let a = input_char !ic in
  let b = input_char !ic in
  let c = input_char !ic in
  let d = input_char !ic in
  let e = input_char !ic in
  let f = input_char !ic in
  let g = input_char !ic in
  let h = input_char !ic in
  float_of_fp8 (mk_fp8 (a, b, c, d, e, f, g, h))

let read_string () =
  let id = read_uint () in
  match !strings with
    | Some tbl -> tbl.(id)
    | None -> assert false

let really_read_string () =
  let len = read_uint () in
  let pad_len = 4 - len mod 4 in
  let total_len = if pad_len < 4 then len + pad_len else len in
  let buf = String.create total_len in
  really_input !ic buf 0 total_len;
  String.sub buf 0 len

let read_array reader =
  let len = read_uint () in
  let rec read_array_aux acc lst =
    if acc = len then lst else read_array_aux (succ acc) (reader () :: lst)
  in
  List.rev (read_array_aux 0 [])

let read_option reader = match read_int () with
  | 0 -> None
  | 1 -> Some (reader ())
  | _ -> assert false

(****************************************************************************
 * READERS FOR THE C AST TYPES (AS DEFINED IN C_AST.X)                      *
 ***************************************************************************)

let rec read_file () =
  let name = read_string () in
  let globals = read_array read_global in
  {
    Cil.fileName = name;
    Cil.globals = globals;
    Cil.globinit = None;
    Cil.globinitcalled = false;
  }

and read_global () = match read_int () with
  | 0 ->
      let name = read_string () in
      let used = read_bool () in
      let c_type = read_c_type () in
      let loc = read_location () in
      let typedef = {
        Cil.tname = name;
        Cil.ttype = c_type;
        Cil.treferenced = used;
      } in
      assert (not (Hashtbl.mem types name));
      Hashtbl.add types name typedef;
      Cil.GType (typedef, loc)
  | 1 ->
      let comp_name = read_string () in
      let kind = match read_int () with
        | 0 -> true (* struct *)
        | 1 -> false (* union *)
        | _ -> assert false
      in
      let key = read_int () in
      let used = read_bool () in
      (* Make a temporary compinfo that the fields can reference. *)
      let comp = {
        Cil.cstruct = kind;
        Cil.cname = comp_name;
        Cil.ckey = key;
        Cil.cfields = []; (* dummy *)
        Cil.cattr = []; (* dummy *)
        Cil.cdefined = true;
        Cil.creferenced = used;
      } in
      (* Put it in the hash table in case the field types need it. *)
      Hashtbl.replace comps comp_name comp;
      let read_field_decl () =
        let field_name = read_string () in
        let width = read_option read_int in
        let c_type = read_c_type () in
        let attrs = read_array read_attribute in
        let loc = read_location () in
        let fi = {
          Cil.fcomp = comp;
          Cil.fname = field_name;
          Cil.ftype = c_type;
          Cil.fbitfield = width;
          Cil.fattr = attrs;
          Cil.floc = loc;
        } in
        Hashtbl.add fields (comp_name, field_name) fi;
        fi
      in
      let fields = read_array read_field_decl in
      let attrs = read_array read_attribute in
      let loc = read_location () in
      (* Now that we have all the fields, put the real list in. *)
      comp.Cil.cfields <- fields;
      comp.Cil.cattr <- attrs;
      Hashtbl.replace comps comp_name comp;
      Cil.GCompTag (comp, loc)
  | 2 ->
      let name = read_string () in
      let kind = match read_int () with
        | 0 -> true (* struct *)
        | 1 -> false (* union *)
        | _ -> assert false
      in
      let key = read_int () in
      let defined = read_bool () in
      let used = read_bool () in
      let attrs = read_array read_attribute in
      let loc = read_location () in
      let comp = {
        Cil.cstruct = kind;
        Cil.cname = name;
        Cil.ckey = key;
        Cil.cfields = [];
        Cil.cattr = attrs;
        Cil.cdefined = defined;
        Cil.creferenced = used;
      } in
      if not (Hashtbl.mem comps name) then Hashtbl.add comps name comp;
      Cil.GCompTagDecl (comp, loc)
  | 3 ->
      let name = read_string () in
      let used = read_bool () in
      let read_enum_item () =
        let name = read_string () in
        let value = read_expr () in
        let loc = read_location () in
        (name, value, loc)
      in
      let items = read_array read_enum_item in
      let attrs = read_array read_attribute in
      let loc = read_location () in
      let enum = {
        Cil.ename = name;
        Cil.eitems = items;
        Cil.eattr = attrs;
        Cil.ereferenced = used;
      } in
      Hashtbl.replace enums name enum;
      Cil.GEnumTag (enum, loc)
  | 4 ->
      let name = read_string () in
      let used = read_bool () in
      let attrs = read_array read_attribute in
      let loc = read_location () in
      let enum = {
        Cil.ename = name;
        Cil.eitems = [];
        Cil.eattr = attrs;
        Cil.ereferenced = used;
      } in
      if not (Hashtbl.mem enums name) then Hashtbl.add enums name enum;
      Cil.GEnumTagDecl (enum, loc)
  | 5 ->
      let name = read_string () in
      let id = read_int () in
      let addr_taken = read_bool () in
      let used = read_bool () in
      let c_type = read_c_type () in
      let stor = read_storage_class () in
      let attrs = read_array read_attribute in
      let loc = read_location () in
      let var = {
        Cil.vname = name;
        Cil.vtype = c_type;
        Cil.vattr = attrs;
        Cil.vstorage = stor;
        Cil.vglob = true;
        Cil.vinline = false;
        Cil.vdecl = loc;
        Cil.vid = id;
        Cil.vaddrof = addr_taken;
        Cil.vreferenced = used;
        Cil.vdescr = Pretty.nil;
        Cil.vdescrpure = false;
      } in
      Hashtbl.replace vars id var;
      let init = read_option read_init in
      let ii = { Cil.init = init } in
      Cil.GVar (var, ii, loc)
  | 6 ->
      let vi = read_var_decl () in
      let loc = vi.Cil.vdecl in
      Cil.GVarDecl (vi, loc)
  | 7 ->
      let name = read_string () in
      let id = read_int () in
      let inline = read_bool () in
      let addr_taken = read_bool () in
      let used = read_bool () in
      let c_type = read_c_type () in
      let stor = read_storage_class () in
      let formals = read_array read_var_decl in
      let locals = read_array read_var_decl in
      (* Of course, since the body might call this function recursively, we
       * must add the varinfo to the hash table first. *)
      let var = {
        Cil.vname = name;
        Cil.vtype = c_type;
        Cil.vattr = []; (* dummy *)
        Cil.vstorage = stor;
        Cil.vglob = true;
        Cil.vinline = inline;
        (* dummy *)
        Cil.vdecl = { Cil.line = -1; Cil.file = ""; Cil.byte = -1; };
        Cil.vid = id;
        Cil.vaddrof = addr_taken;
        Cil.vreferenced = used;
        Cil.vdescr = Pretty.nil;
        Cil.vdescrpure = false;
      } in
      Hashtbl.replace vars id var;
      let body = read_block () in
      let attrs = read_array read_attribute in
      let loc = read_location () in
      var.Cil.vattr <- attrs;
      var.Cil.vdecl <- loc;
      Cil.GFun (
        {
          Cil.svar = var;
          Cil.sformals = formals;
          Cil.slocals = locals;
          Cil.smaxid = -1;
          Cil.sbody = body;
          Cil.smaxstmtid = None;
          Cil.sallstmts = [];
        },
        loc
      )
  | 8 ->
      let name = read_string () in
      let id = read_int () in
      let inline = read_bool () in
      let addr_taken = read_bool () in
      let used = read_bool () in
      let c_type = read_c_type () in
      let stor = read_storage_class () in
      let attrs = read_array read_attribute in
      let loc = read_location () in
      let var = {
        Cil.vname = name;
        Cil.vtype = c_type;
        Cil.vattr = attrs;
        Cil.vstorage = stor;
        Cil.vglob = true;
        Cil.vinline = inline;
        Cil.vdecl = loc;
        Cil.vid = id;
        Cil.vaddrof = addr_taken;
        Cil.vreferenced = used;
        Cil.vdescr = Pretty.nil;
        Cil.vdescrpure = false;
      } in
      if not (Hashtbl.mem vars id) then Hashtbl.add vars id var;
      Cil.GVarDecl (var, loc)
  | 9 ->
      let tmpl = read_string () in
      let loc = read_location () in
      Cil.GAsm (tmpl, loc)
  | 10 ->
      let attr = read_attribute () in
      let loc = read_location () in
      Cil.GPragma (attr, loc)
  | 11 ->
      let s = read_string () in
      Cil.GText s
  | _ -> assert false

and read_storage_class () = match read_int () with
  | 0 -> Cil.NoStorage
  | 1 -> Cil.Static
  | 2 -> Cil.Register
  | 3 -> Cil.Extern
  | _ -> assert false

and read_var_decl () =
  let name = read_string () in
  let id = read_int () in
  let global = read_bool () in
  let addr_taken = read_bool () in
  let used = read_bool () in
  let c_type = read_c_type () in
  let stor = read_storage_class () in
  let attrs = read_array read_attribute in
  let loc = read_location () in
  let var = {
    Cil.vname = name;
    Cil.vtype = c_type;
    Cil.vattr = attrs;
    Cil.vstorage = stor;
    Cil.vglob = global;
    Cil.vinline = false;
    Cil.vdecl = loc;
    Cil.vid = id;
    Cil.vaddrof = addr_taken;
    Cil.vreferenced = used;
    Cil.vdescr = Pretty.nil;
    Cil.vdescrpure = false;
  } in
  if not (Hashtbl.mem vars id) then Hashtbl.add vars id var;
  var

and read_block () =
  let attrs = read_array read_attribute in
  let stmts = read_array read_stmt in
  {
    Cil.battrs = attrs;
    Cil.bstmts = stmts;
  }

and read_int_kind () = match read_int () with
  | 0 -> Cil.IChar
  | 1 -> Cil.ISChar
  | 2 -> Cil.IUChar
  | 3 -> Cil.IInt
  | 4 -> Cil.IUInt
  | 5 -> Cil.IShort
  | 6 -> Cil.IUShort
  | 7 -> Cil.ILong
  | 8 -> Cil.IULong
  | 9 -> Cil.ILongLong
  | 10 -> Cil.IULongLong
  | _ -> assert false

and read_float_kind () = match read_int () with
  | 0 -> Cil.FFloat
  | 1 -> Cil.FDouble
  | 2 -> Cil.FLongDouble
  | _ -> assert false

and read_c_type () = match read_int () with
  | 0 ->
      let attrs = read_array read_attribute in
      Cil.TVoid attrs
  | 1 ->
      let ik = read_int_kind () in
      let attrs = read_array read_attribute in
      Cil.TInt (ik, attrs)
  | 2 ->
      let fk = read_float_kind () in
      let attrs = read_array read_attribute in
      Cil.TFloat (fk, attrs)
  | 3 ->
      let t = read_c_type () in
      let attrs = read_array read_attribute in
      Cil.TPtr (t, attrs)
  | 4 ->
      let t = read_c_type () in
      let len_opt = read_option read_expr in
      let attrs = read_array read_attribute in
      Cil.TArray (t, len_opt, attrs)
  | 5 ->
      let varargs = read_bool () in
      let result = read_c_type () in
      let read_parameter () =
        let t = read_c_type () in
        let attrs = read_array read_attribute in
        ("", t, attrs)
      in
      let formals =
        begin match read_int () with
          | 0 -> Some []
          | 1 -> Some (read_array read_parameter)
          | 2 -> None
          | _ -> assert false
        end
      in
      let attrs = read_array read_attribute in
      Cil.TFun (result, formals, varargs, attrs)
  | 6 ->
      let name = read_string () in
      let attrs = read_array read_attribute in
      Cil.TNamed (Hashtbl.find types name, attrs)
  | 7 ->
      let name = read_string () in
      let attrs = read_array read_attribute in
      (* Not sure if this is stupid hack is a good idea... *)
      begin
        try Cil.TComp (Hashtbl.find comps name, attrs)
        with Not_found -> begin
          let comp = {
            Cil.cstruct = false;
            Cil.cname = name;
            Cil.ckey = -1;
            Cil.cfields = []; (* dummy *)
            Cil.cattr = []; (* dummy *)
            Cil.cdefined = false;
            Cil.creferenced = true;
          } in
          Cil.TComp (comp, attrs)
        end
      end
  | 8 ->
      let name = read_string () in
      let attrs = read_array read_attribute in
      Cil.TEnum (Hashtbl.find enums name, attrs)
  | 9 ->
      let attrs = read_array read_attribute in
      Cil.TBuiltin_va_list attrs
  | _ -> assert false

and read_attribute () =
  let name = read_string () in
  let params = read_array read_attr_param in
  Cil.Attr (name, params)

and read_attr_param () = match read_int () with
  | 0 -> Cil.AInt (read_int ())
  | 1 -> Cil.AStr (read_string ())
  | 2 ->
      let name = read_string () in
      let params = read_array read_attr_param in
      Cil.ACons (name, params)
  | 3 -> Cil.ASizeOf (read_c_type ())
  | 4 -> Cil.ASizeOfE (read_attr_param ())
  | 5 -> Cil.AAlignOf (read_c_type ())
  | 6 -> Cil.AAlignOfE (read_attr_param ())
  | 7 ->
      let op = read_unary_op_kind () in
      let param = read_attr_param () in
      Cil.AUnOp (op, param)
  | 8 ->
      let op = read_binary_op_kind () in
      let param1 = read_attr_param () in
      let param2 = read_attr_param () in
      Cil.ABinOp (op, param1, param2)
  | 9 ->
      let param = read_attr_param () in
      let name = read_string () in
      Cil.ADot (param, name)
  | 10 -> Cil.AStar (read_attr_param ())
  | 11 -> Cil.AAddrOf (read_attr_param ())
  | 12 ->
      let base = read_attr_param () in
      let offset = read_attr_param () in
      Cil.AIndex (base, offset)
  | 13 ->
      let guard = read_attr_param () in
      let tbranch = read_attr_param () in
      let fbranch = read_attr_param () in
      Cil.AQuestion (guard, tbranch, fbranch)
  | _ -> assert false

and read_expr () = match read_int () with
  | 0 -> Cil.Const (read_constant ())
  | 1 -> Cil.Lval (read_lval ())
  | 2 -> Cil.SizeOf (read_c_type ())
  | 3 -> Cil.SizeOfE (read_expr ())
  | 4 -> Cil.SizeOfStr (read_string ())
  | 5 -> Cil.AlignOf (read_c_type ())
  | 6 -> Cil.AlignOfE (read_expr ())
  | 7 ->
      let op = read_unary_op_kind () in
      let e = read_expr () in
      let t = read_c_type () in
      Cil.UnOp (op, e, t)
  | 8 ->
      let op = read_binary_op_kind () in
      let e1 = read_expr () in
      let e2 = read_expr () in
      let t = read_c_type () in
      Cil.BinOp (op, e1, e2, t)
  | 9 ->
      let t = read_c_type () in
      let e = read_expr () in
      Cil.CastE (t, e)
  | 10 -> Cil.AddrOf (read_lval ())
  | 11 -> Cil.StartOf (read_lval ())
  | _ -> assert false

and read_constant () = match read_int () with
  | 0 ->
      let i64 = read_hyper () in
      let ik = read_int_kind () in
      let text_rep_opt = read_option read_string in
      Cil.CInt64 (i64, ik, text_rep_opt)
  | 1 -> Cil.CStr (read_string ())
  | 2 -> Cil.CWStr (read_array read_hyper)
  | 3 -> Cil.CChr (char_of_int (read_int ()))
  | 4 ->
      let f = read_double () in
      let fk = read_float_kind () in
      let text_rep_opt = read_option read_string in
      Cil.CReal (f, fk, text_rep_opt)
  | 5 ->
      let name = read_string () in
      let value = read_expr () in
      let enum_name = read_string () in
      let ei = Hashtbl.find enums enum_name in
      Cil.CEnum (value, name, ei)
  | _ -> assert false

and read_unary_op_kind () = match read_int () with
  | 0 -> Cil.Neg
  | 1 -> Cil.BNot
  | 2 -> Cil.LNot
  | _ -> assert false

and read_binary_op_kind () = match read_int () with
  | 0 -> Cil.PlusA
  | 1 -> Cil.PlusPI
  | 2 -> Cil.IndexPI
  | 3 -> Cil.MinusA
  | 4 -> Cil.MinusPI
  | 5 -> Cil.MinusPP
  | 6 -> Cil.Mult
  | 7 -> Cil.Div
  | 8 -> Cil.Mod
  | 9 -> Cil.Shiftlt
  | 10 -> Cil.Shiftrt
  | 11 -> Cil.Lt
  | 12 -> Cil.Gt
  | 13 -> Cil.Le
  | 14 -> Cil.Ge
  | 15 -> Cil.Eq
  | 16 -> Cil.Ne
  | 17 -> Cil.BAnd
  | 18 -> Cil.BXor
  | 19 -> Cil.BOr
  | 20 -> Cil.LAnd
  | 21 -> Cil.LOr
  | _ -> assert false

and read_lval () =
  let host =
    begin match read_int () with
      | 0 ->
          let id = read_int () in
          let vi = Hashtbl.find vars id in
          Cil.Var vi
      | 1 -> Cil.Mem (read_expr ())
      | _ -> assert false
    end
  in
  let offset = read_offset () in
  (host, offset)

and read_offset () = match read_int () with
  | 0 -> Cil.NoOffset
  | 1 ->
      let field_name = read_string () in
      let comp_name = read_string () in
      let offset = read_offset () in
      Cil.Field (Hashtbl.find fields (comp_name, field_name), offset)
  | 2 ->
      let e = read_expr () in
      let offset = read_offset () in
      Cil.Index (e, offset)
  | _ -> assert false

and read_init () = match read_int () with
  | 0 -> Cil.SingleInit (read_expr ())
  | 1 ->
      let t = read_c_type () in
      let read_init_with_ofst () =
        let o = read_offset () in
        let i = read_init () in
        (o, i)
      in
      let inits = read_array read_init_with_ofst in
      Cil.CompoundInit (t, inits)
  | _ -> assert false

and read_label () = match read_int () with
  | 0 ->
      let s = read_string () in
      let loc = read_location () in
      let in_orig_src = read_bool () in
      Cil.Label (s, loc, in_orig_src)
  | 1 ->
      let e = read_expr () in
      let loc = read_location () in
      Cil.Case (e, loc)
  | 2 ->
      Cil.Default (read_location ())
  | _ -> assert false

and read_stmt () =
  let labels = read_array read_label in
  let is_real_label = function
    | Cil.Label (_, _, _) -> true
    | _ -> false
  in
  let has_real_label = List.exists is_real_label labels in
  (* If the statement we're about to read has a real label (i.e., it can be
   * the target of a goto), then the situation is more complex. *)
  if has_real_label then begin
    let first_label = List.find is_real_label labels in
    let stmt =
      (* Maybe the stmt is already in the hash table because it had to be
       * created so that a goto could reference it. If so, don't create a new
       * stmt; update the fields of that existing one. *)
      if Hashtbl.mem stmts first_label then Hashtbl.find stmts first_label
      (* Otherwise, we need to create a new stmt. We must do so now, despite
       * not having much of the data, because the stmt has to go into the
       * hash table now, since the very content read by read_real_stmt might
       * be a goto referring to this stmt itself. No one with half a brain
       * would ever do such a thing, but it's handled for completeness. *)
      else begin
        let dummy_stmt = {
          Cil.labels = labels;
          (* The rest of these values are placeholders, updated below. *)
          Cil.skind = Cil.Instr [];
          Cil.sid = -1;
          Cil.succs = [];
          Cil.preds = [];
        } in
        Hashtbl.add stmts first_label dummy_stmt;
        dummy_stmt
      end
    in
    (* Now we can read the rest of the content and fill in the stmt. *)
    let real_stmt = read_real_stmt () in
    let id = read_int () in
    let succs = read_array read_stmt in
    let preds = read_array read_stmt in
    stmt.Cil.labels <- labels;
    stmt.Cil.skind <- real_stmt;
    stmt.Cil.sid <- id;
    stmt.Cil.succs <- succs;
    stmt.Cil.preds <- preds;
    stmt
  (* If there are no real labels in the list, then this statement can't be
   * the target of a goto, and the case is much simpler. *)
  end else begin
    let real_stmt = read_real_stmt () in
    let id = read_int () in
    let succs = read_array read_stmt in
    let preds = read_array read_stmt in
    {
      Cil.labels = labels;
      Cil.skind = real_stmt;
      Cil.sid = id;
      Cil.succs = succs;
      Cil.preds = preds;
    }
  end

and read_real_stmt () = match read_int () with
  | 0 -> Cil.Instr (read_array read_instr)
  | 1 ->
      let value_opt = read_option read_expr in
      let loc = read_location () in
      Cil.Return (value_opt, loc)
  | 2 ->
      let label = read_label () in
      let stmt =
        (* If the target statement of the goto, which is identified by the
         * label we just read in, is in the hash table already, then just
         * refer to it. *)
        if Hashtbl.mem stmts label then
          Hashtbl.find stmts label
        (* However, the target might be a statement that hasn't been read in
         * yet. In that case, create a dummy statement that contains no true
         * information and add it to the hash table with the label as its
         * key. When the statement is read in by read_stmt later, instead of
         * creating a new stmt, this dummy one will be retrieved and its
         * fields updated to their true values. It's a hack, but it works. *)
        else begin
          let dummy_stmt = {
            Cil.labels = [];
            Cil.skind = Cil.Instr [];
            Cil.sid = -1;
            Cil.succs = [];
            Cil.preds = [];
          } in
          Hashtbl.add stmts label dummy_stmt;
          dummy_stmt
        end
      in
      let loc = read_location () in
      Cil.Goto (ref stmt, loc)
  | 3 -> Cil.Break (read_location ())
  | 4 -> Cil.Continue (read_location ())
  | 5 ->
      let guard = read_expr () in
      let then_branch = read_block () in
      let else_branch = read_block () in
      let loc = read_location () in
      Cil.If (guard, then_branch, else_branch, loc)
  | 6 ->
      let guard = read_expr () in
      let block = read_block () in
      let cases = read_array read_stmt in
      let loc = read_location () in
      Cil.Switch (guard, block, cases, loc)
  | 7 ->
      let body = read_block () in
      let continue_opt = read_option read_stmt in
      let break_opt = read_option read_stmt in
      let loc = read_location () in
      Cil.Loop (body, loc, continue_opt, break_opt)
  | 8 -> Cil.Block (read_block ())
  | 9 ->
      let try_block = read_block () in
      let finally_block = read_block () in
      let loc = read_location () in
      Cil.TryFinally (try_block, finally_block, loc)
  | 10 ->
      let try_block = read_block () in
      let discards = read_array read_instr in
      let value = read_expr () in
      let except_block = read_block () in
      let loc = read_location () in
      Cil.TryExcept (try_block, (discards, value), except_block, loc)
  | _ -> assert false

and read_instr () = match read_int () with
  | 0 ->
      let lvalue = read_lval () in
      let contents = read_expr () in
      let loc = read_location () in
      Cil.Set (lvalue, contents, loc)
  | 1 ->
      let maybe_result_lvalue = read_option read_lval in
      let name = read_expr () in
      let args = read_array read_expr in
      let loc = read_location () in
      Cil.Call (maybe_result_lvalue, name, args, loc)
  | 2 ->
      let attrs =
        if read_bool () then Cil.Attr ("volatile", []) :: [] else []
      in
      let tmpls = read_array read_string in
      let read_out () =
        let maybe_name = read_option read_string in
        let constrt = read_string () in
        let dest = read_lval () in
        (maybe_name, constrt, dest)
      in
      let outs = read_array read_out in
      let read_in () =
        let maybe_name = read_option read_string in
        let constrt = read_string () in
        let loaded_input = read_expr () in
        (maybe_name, constrt, loaded_input)
      in
      let ins = read_array read_in in
      let clobs = read_array read_string in
      let loc = read_location () in
      Cil.Asm (attrs, tmpls, outs, ins, clobs, loc)
  | _ -> assert false

and read_location () =
  let line = read_int () in
  let file = read_string () in
  let byte = read_int () in
  {
    Cil.line = line;
    Cil.file = file;
    Cil.byte = byte;
  }

(****************************************************************************
 * READER FOR THE TABLE OF INTERNED STRINGS (AS DEFINED IN STRING_TABLE.X)  *
 ***************************************************************************)

let read_string_table () =
  let len = read_uint () in
  strings := Some (Array.make len "");
  let rec read_table_entry acc =
    if acc < len then begin
      let s = really_read_string () in
      let id = read_uint () in
      begin match !strings with
        | Some tbl -> tbl.(id) <- s
        | None -> assert false
      end;
      read_table_entry (succ acc)
    end
  in
  read_table_entry 0

let () =
  let argc = Array.length Sys.argv in
  if argc <> 3 then begin
    eprintf "Usage: %s <AST FILE> <STRINGS FILE>\n" Sys.argv.(0);
    exit 1
  end;
  let string_table_file = Sys.argv.(2) in
  ic := open_in_bin string_table_file;
  read_string_table ();
  let ast_file = Sys.argv.(1) in
  ic := open_in_bin ast_file;
  ignore (read_file ())
