# Readers for the C AST types. A traversal facility for Visitors is also
# provided.

module AST

  class File
    attr_accessor :name, :globals
    protected :name=, :globals=

    def File.read
      file = File.new
      file.name = XDR::String.read
      file.globals = XDR::Array.read Global
      file
    end
    
    def accept(v)
      v.visit_file self
      globals.each { |g| g.accept v }
    end
  end

  class Global
    GLOBAL_TYPE_DEFN = 0
    GLOBAL_COMPOSITE_TAG_DEFN = 1
    GLOBAL_COMPOSITE_TAG_DECL = 2
    GLOBAL_ENUM_TAG_DEFN = 3
    GLOBAL_ENUM_TAG_DECL = 4
    GLOBAL_VARIABLE_DEFN = 5
    GLOBAL_VARIABLE_DECL = 6
    GLOBAL_FUNCTION_DEFN = 7
    GLOBAL_FUNCTION_DECL = 8
    GLOBAL_ASM_STATEMENT = 9
    GLOBAL_PRAGMA = 10
    GLOBAL_TEXT = 11

    def Global.read
      case XDR::Int.read
      when GLOBAL_TYPE_DEFN: TypeDefn.read
      when GLOBAL_COMPOSITE_TAG_DEFN: CompDefn.read
      when GLOBAL_COMPOSITE_TAG_DECL: CompDecl.read
      when GLOBAL_ENUM_TAG_DEFN: EnumDefn.read
      when GLOBAL_ENUM_TAG_DECL: EnumDecl.read
      when GLOBAL_VARIABLE_DEFN: VarDefn.read
      when GLOBAL_VARIABLE_DECL: VarDecl.read
      when GLOBAL_FUNCTION_DEFN: FuncDefn.read
      when GLOBAL_FUNCTION_DECL: FuncDecl.read
      when GLOBAL_ASM_STATEMENT: GlobAsm.read
      when GLOBAL_PRAGMA: Pragma.read
      when GLOBAL_TEXT: Text.read
      else raise "Invalid discriminator for Global"
      end
    end
  end

  class TypeDefn < Global
    attr_accessor :name, :used, :type, :loc
    protected :name=, :used=, :type=, :loc=

    def TypeDefn.read
      type_defn = TypeDefn.new
      type_defn.name = XDR::String.read
      type_defn.used = XDR::Bool.read
      type_defn.type = Type.read
      type_defn.loc = Location.read
      type_defn
    end
    
    def accept(v)
      v.visit_global self
      type.accept v
    end
  end

  class CompKind
    COMPOSITE_STRUCT = 0
    COMPOSITE_UNION = 1

    def CompKind.read
      comp_kind = XDR::Int.read
      raise "Invalid discriminator for CompKind" unless (0..1) === comp_kind
      comp_kind
    end
  end

  class FieldDecl
    attr_accessor :name, :bit_width, :type, :attrs, :loc
    protected :name=, :bit_width=, :type=, :attrs=, :loc=

    def FieldDecl.read
      field_decl = FieldDecl.new
      field_decl.name = XDR::String.read
      field_decl.bit_width = XDR::Option.read XDR::Int
      field_decl.type = Type.read
      field_decl.attrs = XDR::Array.read Attribute
      field_decl.loc = Location.read
      field_decl
    end
    
    def accept(v)
      type.accept v
      attrs.each { |a| a.accept v }
    end
  end

  class CompDefn < Global
    attr_accessor :name, :kind, :key, :used, :fields, :attrs, :loc
    protected :name=, :kind=, :key=, :used=, :fields=, :attrs=, :loc=

    def CompDefn.read
      comp_defn = CompDefn.new
      comp_defn.name = XDR::String.read
      comp_defn.kind = CompKind.read
      comp_defn.key = XDR::Int.read
      comp_defn.used = XDR::Bool.read
      comp_defn.fields = XDR::Array.read FieldDecl
      comp_defn.attrs = XDR::Array.read Attribute
      comp_defn.loc = Location.read
      comp_defn
    end
    
    def accept(v)
      v.visit_global self
      fields.each { |f| f.accept v }
      attrs.each { |a| a.accept v }
    end
  end

  class CompDecl < Global
    attr_accessor :name, :kind, :key, :defined, :used, :attrs, :loc
    protected :name=, :kind=, :key=, :defined=, :used=, :attrs=, :loc=

    def CompDecl.read
      comp_decl = CompDecl.new
      comp_decl.name = XDR::String.read
      comp_decl.kind = CompKind.read
      comp_decl.key = XDR::Int.read
      comp_decl.defined = XDR::Bool.read
      comp_decl.used = XDR::Bool.read
      comp_decl.attrs = XDR::Array.read Attribute
      comp_decl.loc = Location.read
      comp_decl
    end
    
    def accept(v)
      v.visit_global self
      attrs.each { |f| f.accept v }
    end
  end

  class EnumItem
    attr_accessor :name, :value, :loc
    protected :name=, :value=, :loc=

    def EnumItem.read
      enum_item = EnumItem.new
      enum_item.name = XDR::String.read
      enum_item.value = Expr.read
      enum_item.loc = Location.read
      enum_item
    end
    
    def accept(v)
      value.accept v
    end
  end

  class EnumDefn < Global
    attr_accessor :name, :used, :items, :attrs, :loc
    protected :name=, :used=, :items=, :attrs=, :loc=

    def EnumDefn.read
      enum_defn = EnumDefn.new
      enum_defn.name = XDR::String.read
      enum_defn.used = XDR::Bool.read
      enum_defn.items = XDR::Array.read EnumItem
      enum_defn.attrs = XDR::Array.read Attribute
      enum_defn.loc = Location.read
      enum_defn
    end
    
    def accept(v)
      v.visit_global self
      items.each { |i| i.accept v }
      attrs.each { |a| a.accept v }
    end
  end

  class EnumDecl < Global
    attr_accessor :name, :used, :attrs, :loc
    protected :name=, :used=, :attrs=, :loc=

    def EnumDecl.read
      enum_decl = EnumDecl.new
      enum_decl.name = XDR::String.read
      enum_decl.used = XDR::Bool.read
      enum_decl.attrs = XDR::Array.read Attribute
      enum_decl.loc = Location.read
      enum_decl
    end
    
    def accept(v)
      v.visit_global self
      attrs.each { |a| a.accept v }
    end
  end

  class StorageClass
    NO_STORAGE = 0
    STATIC = 1
    REGISTER = 2
    EXTERN = 3

    def StorageClass.read
      storage_class = XDR::Int.read
      unless (0..3) === storage_class
        raise "Invalid discriminator for StorageClass"
      end
      storage_class
    end
  end

  class VarDefn < Global
    attr_accessor :name, :id, :address_taken, :used, :type, :storage, :attrs
    attr_accessor :loc, :initializer
    protected :name=, :id=, :address_taken=, :used=, :type=, :storage=
    protected :attrs=, :loc=, :initializer=

    def VarDefn.read
      var_defn = VarDefn.new
      var_defn.name = XDR::String.read
      var_defn.id = XDR::Int.read
      var_defn.address_taken = XDR::Bool.read
      var_defn.used = XDR::Bool.read
      var_defn.type = Type.read
      var_defn.storage = StorageClass.read
      var_defn.attrs = XDR::Array.read Attribute
      var_defn.loc = Location.read
      var_defn.initializer = XDR::Option.read Init
      var_defn
    end
    
    def accept(v)
      v.visit_global self
      type.accept v
      attrs.each { |a| a.accept v }
      initializer.accept v if initializer
    end
  end

  class VarDecl < Global
    attr_accessor :name, :id, :global, :address_taken, :used, :type, :storage
    attr_accessor :attrs, :loc
    protected :name=, :id=, :global=, :address_taken=, :used=, :type=
    protected :storage=, :attrs=, :loc=

    def VarDecl.read
      var_decl = VarDecl.new
      var_decl.name = XDR::String.read
      var_decl.id = XDR::Int.read
      var_decl.global = XDR::Bool.read
      var_decl.address_taken = XDR::Bool.read
      var_decl.used = XDR::Bool.read
      var_decl.type = Type.read
      var_decl.storage = StorageClass.read
      var_decl.attrs = XDR::Array.read Attribute
      var_decl.loc = Location.read
      var_decl
    end
    
    def accept(v)
      global ? v.visit_global(self) : v.visit_local_var_decl(self)
      type.accept v
      attrs.each { |a| a.accept v }
    end
  end

  class Block
    attr_accessor :attrs, :stmts
    protected :attrs=, :stmts=

    def Block.read
      block = Block.new
      block.attrs = XDR::Array.read Attribute
      block.stmts = XDR::Array.read Stmt
      block
    end
    
    def accept(v)
      attrs.each { |a| a.accept v }
      stmts.each { |s| s.accept v }
    end
  end

  class FuncDefn < Global
    attr_accessor :name, :id, :inline, :address_taken, :used, :type, :storage
    attr_accessor :params, :locals, :body, :attrs, :loc
    protected :name=, :id=, :inline=, :address_taken=, :used=, :type=
    protected :storage=, :params=, :locals=, :body=, :attrs=, :loc=

    def FuncDefn.read
      func_defn = FuncDefn.new
      func_defn.name = XDR::String.read
      func_defn.id = XDR::Int.read
      func_defn.inline = XDR::Bool.read
      func_defn.address_taken = XDR::Bool.read
      func_defn.used = XDR::Bool.read
      func_defn.type = Type.read
      func_defn.storage = StorageClass.read
      func_defn.params = XDR::Array.read VarDecl
      func_defn.locals = XDR::Array.read VarDecl
      func_defn.body = Block.read
      func_defn.attrs = XDR::Array.read Attribute
      func_defn.loc = Location.read
      $funcs[func_defn.id] = func_defn
    end
    
    def accept(v)
      v.visit_global self
      type.accept v
      params.each { |p| p.accept v }
      locals.each { |l| l.accept v }
      body.accept v
      attrs.each { |a| a.accept v }
    end
  end

  class FuncDecl < Global
    attr_accessor :name, :id, :inline, :address_taken, :used, :type, :storage
    attr_accessor :attrs, :loc
    protected :name=, :id=, :inline=, :address_taken=, :used=, :type=
    protected :storage=, :attrs=, :loc=

    def FuncDecl.read
      func_decl = FuncDecl.new
      func_decl.name = XDR::String.read
      func_decl.id = XDR::Int.read
      func_decl.inline = XDR::Bool.read
      func_decl.address_taken = XDR::Bool.read
      func_decl.used = XDR::Bool.read
      func_decl.type = Type.read
      func_decl.storage = StorageClass.read
      func_decl.attrs = XDR::Array.read Attribute
      func_decl.loc = Location.read
      $funcs[func_decl.id] = func_decl
    end
    
    def accept(v)
      v.visit_global self
      type.accept v
      attrs.each { |a| a.accept v }
    end
  end

  class GlobAsm < Global
    attr_accessor :template, :loc
    protected :template=, :loc=

    def GlobAsm.read
      glob_asm = GlobAsm.new
      glob_asm.template = XDR::String.read
      glob_asm.loc = Location.read
      glob_asm
    end
    
    def accept(v)
      v.visit_global self
    end
  end

  class Pragma < Global
    attr_accessor :content, :loc
    protected :content=, :loc=

    def Pragma.read
      pragma = Pragma.new
      pragma.content = Attribute.read
      pragma.loc = Location.read
      pragma
    end
    
    def accept(v)
      v.visit_global self
    end
  end

  class Text < Global
    attr_accessor :str
    protected :str=

    def Text.read
      text = Text.new
      text.str = XDR::String.read
      text
    end
    
    def accept(v)
      v.visit_global self
    end
  end
  
  class Type
    TYPE_VOID = 0
    TYPE_INT = 1
    TYPE_FLOAT = 2
    TYPE_POINTER = 3
    TYPE_ARRAY = 4
    TYPE_FUNCTION = 5
    TYPE_NAMED = 6
    TYPE_COMPOSITE = 7
    TYPE_ENUMERATED = 8
    TYPE_GCC_BUILTIN_VA_LIST = 9

    def Type.read
      case XDR::Int.read
      when TYPE_VOID: Void.read
      when TYPE_INT: Int.read
      when TYPE_FLOAT: Float.read
      when TYPE_POINTER: Ptr.read
      when TYPE_ARRAY: Array.read
      when TYPE_FUNCTION: Fun.read
      when TYPE_NAMED: Named.read
      when TYPE_COMPOSITE: Comp.read
      when TYPE_ENUMERATED: Enum.read
      when TYPE_GCC_BUILTIN_VA_LIST: GCCBuiltinVAList.read
      else raise "Invalid discriminator for Type"
      end
    end
  end

  class Void < Type
    attr_accessor :attrs
    protected :attrs=

    def Void.read
      c_void = Void.new
      c_void.attrs = XDR::Array.read Attribute
      c_void
    end
    
    def accept(v)
      v.visit_type self
      attrs.each { |a| a.accept v }
    end
  end

  class IntKind
    CHAR = 0
    SIGNED_CHAR = 1
    UNSIGNED_CHAR = 2
    INT = 3
    UNSIGNED_INT = 4
    SHORT = 5
    UNSIGNED_SHORT = 6
    LONG = 7
    UNSIGNED_LONG = 8
    LONG_LONG = 9
    UNSIGNED_LONG_LONG = 10

    def IntKind.read
      kind = XDR::Int.read
      raise "Invalid discriminator for IntKind" unless (0..10) === kind
      kind
    end
  end

  class FloatKind
    FLOAT = 0
    DOUBLE = 1
    LONG_DOUBLE = 2

    def FloatKind.read
      kind = XDR::Int.read
      raise "Invalid discriminator for FloatKind" unless (0..2) === kind
      kind
    end
  end

  class Int < Type
    attr_accessor :kind, :attrs
    protected :kind=, :attrs=

    def Int.read
      c_int = Int.new
      c_int.kind = IntKind.read
      c_int.attrs = XDR::Array.read Attribute
      c_int
    end
    
    def accept(v)
      v.visit_type self
      attrs.each { |a| a.accept v }
    end
  end

  class Float < Type
    attr_accessor :kind, :attrs
    protected :kind=, :attrs=

    def Float.read
      c_float = Float.new
      c_float.kind = FloatKind.read
      c_float.attrs = XDR::Array.read Attribute
      c_float
    end
    
    def accept(v)
      v.visit_type self
      attrs.each { |a| a.accept v }
    end
  end

  class Ptr < Type
    attr_accessor :pointed_to, :attrs
    protected :pointed_to=, :attrs=

    def Ptr.read
      c_ptr = Ptr.new
      c_ptr.pointed_to = Type.read
      c_ptr.attrs = XDR::Array.read Attribute
      c_ptr
    end
    
    def accept(v)
      v.visit_type self
      pointed_to.accept v
      attrs.each { |a| a.accept v }
    end
  end

  class Array < Type
    attr_accessor :members, :length, :attrs
    protected :members=, :length=, :attrs=

    def Array.read
      c_array = Array.new
      c_array.members = Type.read
      c_array.length = XDR::Option.read Expr
      c_array.attrs = XDR::Array.read Attribute
      c_array
    end
    
    def accept(v)
      v.visit_type self
      members.accept v
      length.accept v if length
      attrs.each { |a| a.accept v }
    end
  end

  class Parameter
    attr_accessor :type, :attrs
    protected :type=, :attrs=

    def Parameter.read
      parameter = Parameter.new
      parameter.type = Type.read
      parameter.attrs = XDR::Array.read Attribute
      parameter
    end
    
    def accept(v)
      type.accept v
      attrs.each { |a| a.accept v }
    end
  end

  class ParameterList
    NO_PARAMETERS = 0
    SOME_PARAMETERS = 1
    PARAMETERS_UNKNOWN = 2

    def ParameterList.read
      case XDR::Int.read
      when NO_PARAMETERS: []
      when SOME_PARAMETERS: XDR::Array.read Parameter
      when PARAMETERS_UNKNOWN: nil
      else raise "Invalid discriminator for ParameterList"
      end
    end
  end

  class Fun < Type
    attr_accessor :varargs, :result, :formals, :attrs
    protected :varargs=, :result=, :formals=, :attrs=

    def Fun.read
      c_fun = Fun.new
      c_fun.varargs = XDR::Bool.read
      c_fun.result = Type.read
      c_fun.formals = ParameterList.read
      c_fun.attrs = XDR::Array.read Attribute
      c_fun
    end
    
    def accept(v)
      v.visit_type self
      result.accept v
      formals.each { |f| f.accept v if f } if formals
      attrs.each { |a| a.accept v }
    end
  end

  class Named < Type
    attr_accessor :name, :attrs
    protected :name=, :attrs=

    def Named.read
      c_named = Named.new
      c_named.name = XDR::String.read
      c_named.attrs = XDR::Array.read Attribute
      c_named
    end
    
    def accept(v)
      v.visit_type self
      attrs.each { |a| a.accept v }
    end
  end

  class Comp < Type
    attr_accessor :name, :attrs
    protected :name=, :attrs=

    def Comp.read
      c_comp = Comp.new
      c_comp.name = XDR::String.read
      c_comp.attrs = XDR::Array.read Attribute
      c_comp
    end
    
    def accept(v)
      v.visit_type self
      attrs.each { |a| a.accept v }
    end
  end

  class Enum < Type
    attr_accessor :name, :attrs
    protected :name=, :attrs=

    def Enum.read
      c_enum = Enum.new
      c_enum.name = XDR::String.read
      c_enum.attrs = XDR::Array.read Attribute
      c_enum
    end
    
    def accept(v)
      v.visit_type self
      attrs.each { |a| a.accept v }
    end
  end

  class GCCBuiltinVAList < Type
    attr_accessor :attrs
    protected :attrs=

    def GCCBuiltinVAList.read
      va_list = GCCBuiltinVAList.new
      va_list.attrs = XDR::Array.read Attribute
      va_list
    end
    
    def accept(v)
      v.visit_type self
      attrs.each { |a| a.accept v }
    end
  end

  class Attribute
    attr_accessor :name, :params
    protected :name=, :params=

    def Attribute.read
      attribute = Attribute.new
      attribute.name = XDR::String.read
      attribute.params = XDR::Array.read AttrParam
      attribute
    end
    
    def accept(v)
      v.visit_attribute self
      params.each { |p| p.accept v }
    end
  end
  
  class AttrParam
    PARAM_INT = 0
    PARAM_STRING = 1
    PARAM_CONSTRUCTED = 2
    PARAM_SIZEOF_TYPE = 3
    PARAM_SIZEOF_PARAM = 4
    PARAM_SIZEOF_TYPESIG = 5
    PARAM_ALIGNOF_TYPE = 6
    PARAM_ALIGNOF_PARAM = 7
    PARAM_ALIGNOF_TYPESIG = 8
    PARAM_UNARY_OP = 9
    PARAM_BINARY_OP = 10
    PARAM_DOT = 11
    PARAM_STAR = 12
    PARAM_ADDROF = 13
    PARAM_INDEX = 14
    PARAM_TERNARY_OP = 15

    def AttrParam.read
      case XDR::Int.read
      when PARAM_INT: AttrInt.read
      when PARAM_STRING: AttrString.read
      when PARAM_CONSTRUCTED: AttrCons.read
      when PARAM_SIZEOF_TYPE: AttrSizeofType.read
      when PARAM_SIZEOF_PARAM: AttrSizeofParam.read
      when PARAM_SIZEOF_TYPESIG: AttrSizeofTypesig.read
      when PARAM_ALIGNOF_TYPE: AttrAlignofType.read
      when PARAM_ALIGNOF_PARAM: AttrAlignofParam.read
      when PARAM_ALIGNOF_TYPESIG: AttrAlignofTypesig.read
      when PARAM_UNARY_OP: AttrUnary.read
      when PARAM_BINARY_OP: AttrBinary.read
      when PARAM_DOT: AttrDot.read
      when PARAM_STAR: AttrStar.read
      when PARAM_ADDROF: AttrAddrof.read
      when PARAM_INDEX: AttrIndex.read
      when PARAM_TERNARY_OP: AttrTernary.read
      else raise "Invalid discriminator for AttrParam"
      end
    end
  end
  
  class AttrInt < AttrParam
    attr_accessor :int
    protected :int=
    
    def AttrInt.read
      attr_int = AttrInt.new
      attr_int.int = XDR::Int.read
      attr_int
    end
    
    def accept(v)
      v.visit_attr_param self
    end
  end
  
  class AttrString < AttrParam
    attr_accessor :string
    protected :string=
    
    def AttrString.read
      attr_string = AttrString.new
      attr_string.string = XDR::String.read
      attr_string
    end
    
    def accept(v)
      v.visit_attr_param self
    end
  end

  class AttrCons < AttrParam
    attr_accessor :name, :params
    protected :name=, :params=

    def AttrCons.read
      attr_cons = AttrCons.new
      attr_cons.name = XDR::String.read
      attr_cons.params = XDR::Array.read AttrParam
      attr_cons
    end
    
    def accept(v)
      v.visit_attr_param self
      params.each { |p| p.accept v }
    end
  end

  class AttrSizeofType < AttrParam
    attr_accessor :type
    protected :type=

    def AttrSizeofType.read
      attr_sizeof_type = AttrSizeofType.new
      attr_sizeof_type.type = Type.read
      attr_sizeof_type
    end
    
    def accept(v)
      v.visit_attr_param self
      type.accept v
    end
  end

  class AttrSizeofParam < AttrParam
    attr_accessor :param
    protected :param=

    def AttrSizeofParam.read
      attr_sizeof_param = AttrSizeofParam.new
      attr_sizeof_param.param = AttrParam.read
      attr_sizeof_param
    end
    
    def accept(v)
      v.visit_attr_param self
      param.accept v
    end
  end

  class AttrSizeofTypesig < AttrParam
    attr_accessor :typesig
    protected :typesig=

    def AttrSizeofTypesig.read
      attr_sizeof_typesig = AttrSizeofTypesig.new
      attr_sizeof_typesig.typesig = Typesig.read
      attr_sizeof_typesig
    end
    
    def accept(v)
      v.visit_attr_param self
      typesig.accept v
    end
  end

  class AttrAlignofType < AttrParam
    attr_accessor :type
    protected :type=

    def AttrAlignofType.read
      attr_alignof_type = AttrAlignofType.new
      attr_alignof_type.type = Type.read
      attr_alignof_type
    end
    
    def accept(v)
      v.visit_attr_param self
      type.accept v
    end
  end

  class AttrAlignofParam < AttrParam
    attr_accessor :param
    protected :param=

    def AttrAlignofParam.read
      attr_alignof_param = AttrAlignofParam.new
      attr_alignof_param.param = AttrParam.read
      attr_alignof_param
    end
    
    def accept(v)
      v.visit_attr_param self
      param.accept v
    end
  end

  class AttrAlignofTypesig < AttrParam
    attr_accessor :typesig
    protected :typesig=

    def AttrAlignofTypesig.read
      attr_alignof_typesig = AttrAlignofTypesig.new
      attr_alignof_typesig.typesig = Typesig.read
      attr_alignof_typesig
    end
    
    def accept(v)
      v.visit_attr_param self
      typesig.accept v
    end
  end

  class AttrUnary < AttrParam
    attr_accessor :operator, :param
    protected :operator=, :param=

    def AttrUnary.read
      attr_unary = AttrUnary.new
      attr_unary.operator = UnaryOpKind.read
      attr_unary.param = AttrParam.read
      attr_unary
    end
    
    def accept(v)
      v.visit_attr_param self
      param.accept v
    end
  end

  class AttrBinary < AttrParam
    attr_accessor :operator, :first_param, :second_param
    protected :operator=, :first_param=, :second_param=

    def AttrBinary.read
      attr_binary = AttrBinary.new
      attr_binary.operator = BinaryOpKind.read
      attr_binary.first_param = AttrParam.read
      attr_binary.second_param = AttrParam.read
      attr_binary
    end
    
    def accept(v)
      v.visit_attr_param self
      first_param.accept v
      second_param.accept v
    end
  end

  class AttrDot < AttrParam
    attr_accessor :param, :name
    protected :param=, :name=

    def AttrDot.read
      attr_dot = AttrDot.new
      attr_dot.param = AttrParam.read
      attr_dot.name = XDR::String.read
      attr_dot
    end
    
    def accept(v)
      v.visit_attr_param self
      param.accept v
    end
  end

  class AttrStar < AttrParam
    attr_accessor :param
    protected :param=

    def AttrStar.read
      attr_star = AttrStar.new
      attr_star.param = AttrParam.read
      attr_star
    end
    
    def accept(v)
      v.visit_attr_param self
      param.accept v
    end
  end

  class AttrAddrof < AttrParam
    attr_accessor :param
    protected :param=

    def AttrAddrof.read
      attr_addrof = AttrAddrof.new
      attr_addrof.param = AttrParam.read
      attr_addrof
    end
    
    def accept(v)
      v.visit_attr_param self
      param.accept v
    end
  end

  class AttrIndex < AttrParam
    attr_accessor :base, :offset
    protected :base=, :offset=

    def AttrIndex.read
      attr_index = AttrIndex.new
      attr_index.base = AttrParam.read
      attr_index.offset = AttrParam.read
      attr_index
    end
    
    def accept(v)
      v.visit_attr_param self
      base.accept v
      offset.accept v
    end
  end

  class AttrTernary < AttrParam
    attr_accessor :guard, :true_branch, :false_branch
    protected :guard=, :true_branch=, :false_branch=

    def AttrTernary.read
      attr_ternary = AttrTernary.new
      attr_ternary.guard = AttrParam.read
      attr_ternary.true_branch = AttrParam.read
      attr_ternary.false_branch = AttrParam.read
      attr_ternary
    end
    
    def accept(v)
      v.visit_attr_param self
      guard.accept v
      true_branch.accept v
      false_branch.accept v
    end
  end

  class Expr
    EXPR_CONST = 0
    EXPR_LVALUE = 1
    EXPR_SIZEOF_TYPE = 2
    EXPR_SIZEOF_EXPR = 3
    EXPR_SIZEOF_STRING_LITERAL = 4
    EXPR_ALIGNOF_TYPE = 5
    EXPR_ALIGNOF_EXPR = 6
    EXPR_UNARY_OP = 7
    EXPR_BINARY_OP = 8
    EXPR_TYPE_CAST = 9
    EXPR_ADDROF = 10
    EXPR_STARTOF = 11

    def Expr.read
      case discr = XDR::Int.read
      when EXPR_CONST: ExprConstant.read
      when EXPR_LVALUE: ExprLvalue.read
      when EXPR_SIZEOF_TYPE: ExprSizeofType.read
      when EXPR_SIZEOF_EXPR: ExprSizeofExpr.read
      when EXPR_SIZEOF_STRING_LITERAL: ExprSizeofStringLiteral.read
      when EXPR_ALIGNOF_TYPE: ExprAlignofType.read
      when EXPR_ALIGNOF_EXPR: ExprAlignofExpr.read
      when EXPR_UNARY_OP: ExprUnary.read
      when EXPR_BINARY_OP: ExprBinary.read
      when EXPR_TYPE_CAST: ExprCast.read
      when EXPR_ADDROF: ExprAddrof.read
      when EXPR_STARTOF: ExprStartof.read
      else raise "Invalid discriminator for Expr"
      end
    end
    
    def identifier
      "(unknown)"
    end
  end

  class ExprConstant < Expr
    attr_accessor :constant
    protected :constant=

    def ExprConstant.read
      expr_constant = ExprConstant.new
      expr_constant.constant = Constant.read
      expr_constant
    end
    
    def accept(v)
      v.visit_expr self
      constant.accept v
    end
  end

  class ExprLvalue < Expr
    attr_accessor :lvalue
    protected :lvalue=

    def ExprLvalue.read
      expr_lvalue = ExprLvalue.new
      expr_lvalue.lvalue = Lval.read
      expr_lvalue
    end
    
    def accept(v)
      v.visit_expr self
      lvalue.accept v
    end
    
    def identifier
      lvalue.identifier
    end
  end

  class ExprSizeofType < Expr
    attr_accessor :type
    protected :type=

    def ExprSizeofType.read
      expr_sizeof_type = ExprSizeofType.new
      expr_sizeof_type.type = Type.read
      expr_sizeof_type
    end
    
    def accept(v)
      v.visit_expr self
      type.accept v
    end
  end

  class ExprSizeofExpr < Expr
    attr_accessor :expr
    protected :expr=

    def ExprSizeofExpr.read
      expr_sizeof_expr = ExprSizeofExpr.new
      expr_sizeof_expr.expr = Expr.read
      expr_sizeof_expr
    end
    
    def accept(v)
      v.visit_expr self
      expr.accept v
    end
  end

  class ExprSizeofStringLiteral < Expr
    attr_accessor :str
    protected :str=

    def ExprSizeofStringLiteral.read
      expr_sizeof_string_literal = ExprSizeofStringLiteral.new
      expr_sizeof_string_literal.str = XDR::String.read
      expr_sizeof_string_literal
    end
    
    def accept(v)
      v.visit_expr self
    end
  end

  class ExprAlignofType < Expr
    attr_accessor :type
    protected :type=

    def ExprAlignofType.read
      expr_alignof_type = ExprAlignofType.new
      expr_alignof_type.type = Type.read
      expr_alignof_type
    end
    
    def accept(v)
      v.visit_expr self
      type.accept v
    end
  end

  class ExprAlignofExpr < Expr
    attr_accessor :expr
    protected :expr=

    def ExprAlignofExpr.read
      expr_alignof_expr = ExprAlignofExpr.new
      expr_alignof_expr.expr = Expr.read
      expr_alignof_expr
    end
    
    def accept(v)
      v.visit_expr self
      expr.accept v
    end
  end

  class ExprUnary < Expr
    attr_accessor :operator, :operand, :type
    protected :operator=, :operand=, :type=

    def ExprUnary.read
      expr_unary = ExprUnary.new
      expr_unary.operator = UnaryOpKind.read
      expr_unary.operand = Expr.read
      expr_unary.type = Type.read
      expr_unary
    end
    
    def accept(v)
      v.visit_expr self
      operand.accept v
      type.accept v
    end
  end

  class ExprBinary < Expr
    attr_accessor :operator, :first_operand, :second_operand, :type
    protected :operator=, :first_operand=, :second_operand=, :type=

    def ExprBinary.read
      expr_binary = ExprBinary.new
      expr_binary.operator = BinaryOpKind.read
      expr_binary.first_operand = Expr.read
      expr_binary.second_operand = Expr.read
      expr_binary.type = Type.read
      expr_binary
    end
    
    def accept(v)
      v.visit_expr self
      first_operand.accept v
      second_operand.accept v
      type.accept v
    end
  end

  class ExprCast < Expr
    attr_accessor :type, :expr
    protected :type=, :expr=

    def ExprCast.read
      expr_cast = ExprCast.new
      expr_cast.type = Type.read
      expr_cast.expr = Expr.read
      expr_cast
    end
    
    def accept(v)
      v.visit_expr self
      type.accept v
      expr.accept v
    end
  end

  class ExprAddrof < Expr
    attr_accessor :lvalue
    protected :lvalue=

    def ExprAddrof.read
      expr_addrof = ExprAddrof.new
      expr_addrof.lvalue = Lval.read
      expr_addrof
    end
    
    def accept(v)
      v.visit_expr self
      lvalue.accept v
    end
  end

  class ExprStartof < Expr
    attr_accessor :lvalue
    protected :lvalue=

    def ExprStartof.read
      expr_startof = ExprStartof.new
      expr_startof.lvalue = Lval.read
      expr_startof
    end
    
    def accept(v)
      v.visit_expr self
      lvalue.accept v
    end
  end
  
  class Constant
    CONST_INTEGER = 0
    CONST_STRING = 1
    CONST_WIDE_STRING = 2
    CONST_CHAR = 3
    CONST_FLOATING_POINT = 4
    CONST_ENUMERATION = 5

    def Constant.read
      case XDR::Int.read
      when CONST_INTEGER: IntNum.read
      when CONST_STRING: StringConst.read
      when CONST_WIDE_STRING: WideStringConst.read
      when CONST_CHAR: CharConst.read
      when CONST_FLOATING_POINT: FloatNum.read
      when CONST_ENUMERATION: EnumVal.read
      else raise "Invalid discriminator for Constant"
      end
    end
  end

  class IntNum < Constant
    attr_accessor :value, :kind, :text_rep
    protected :value=, :kind=, :text_rep=

    def IntNum.read
      int_num = IntNum.new
      int_num.value = XDR::Hyper.read
      int_num.kind = IntKind.read
      int_num.text_rep = XDR::Option.read XDR::String
      int_num
    end
    
    def accept(v)
      v.visit_constant self
    end
  end
  
  class StringConst < Constant
    attr_accessor :string
    protected :string=
    
    def StringConst.read
      string_const = StringConst.new
      string_const.string = XDR::String.read
      string_const
    end
    
    def accept(v)
      v.visit_constant self
    end
  end
  
  class WideStringConst < Constant
    attr_accessor :wide_string
    protected :wide_string=
    
    def WideStringConst.read
      wide_string_const = WideStringConst.new
      wide_string_const.wide_string = XDR::Array.read XDR::Hyper
      wide_string_const
    end
    
    def accept(v)
      v.visit_constant self
    end
  end
  
  class CharConst < Constant
    attr_accessor :char
    protected :char=
    
    def CharConst.read
      char_const = CharConst.new
      char_const.char = XDR::Int.read
      char_const
    end
    
    def accept(v)
      v.visit_constant self
    end
  end

  class FloatNum < Constant
    attr_accessor :value, :kind, :text_rep
    protected :value=, :kind=, :text_rep=

    def FloatNum.read
      float_num = FloatNum.new
      float_num.value = XDR::Double.read
      float_num.kind = FloatKind.read
      float_num.text_rep = XDR::Option.read XDR::String
      float_num
    end
    
    def accept(v)
      v.visit_constant self
    end
  end

  class EnumVal < Constant
    attr_accessor :name, :value, :host_enum_name
    protected :name=, :value=, :host_enum_name=

    def EnumVal.read
      enum_val = EnumVal.new
      enum_val.name = XDR::String.read
      enum_val.value = Expr.read
      enum_val.host_enum_name = XDR::String.read
      enum_val
    end
    
    def accept(v)
      v.visit_constant self
      value.accept v
    end
  end

  class UnaryOpKind
    UNOP_NEGATION = 0
    UNOP_BITWISE_COMPLEMENT = 1
    UNOP_LOGICAL_NOT = 2

    def UnaryOpKind.read
      unary_op_kind = XDR::Int.read
      unless (0..2) === unary_op_kind
        raise "Invalid discriminator for UnaryOpKind"
      end
      unary_op_kind
    end
  end

  class BinaryOpKind
    BINOP_ARITHMETIC_PLUS = 0
    BINOP_POINTER_PLUS_INTEGER = 1
    BINOP_POINTER_PLUS_INTEGER_FROM_ARRAY = 2
    BINOP_ARITHMETIC_MINUS = 3
    BINOP_POINTER_MINUS_INTEGER = 4
    BINOP_POINTER_MINUS_POINTER = 5
    BINOP_MULTIPLICATION = 6
    BINOP_DIVISION = 7
    BINOP_MODULUS = 8
    BINOP_SHIFT_LEFT = 9
    BINOP_SHIFT_RIGHT = 10
    BINOP_LESS_THAN = 11
    BINOP_GREATER_THAN = 12
    BINOP_LESS_OR_EQUAL = 13
    BINOP_GREATER_OR_EQUAL = 14
    BINOP_EQUAL = 15
    BINOP_NOT_EQUAL = 16
    BINOP_BITWISE_AND = 17
    BINOP_BITWISE_XOR = 18
    BINOP_BITWISE_OR = 19
    BINOP_LOGICAL_AND = 20
    BINOP_LOGICAL_OR = 21

    def BinaryOpKind.read
      binary_op_kind = XDR::Int.read
      unless (0..21) === binary_op_kind
        raise "Invalid discriminator for BinaryOpKind"
      end
      binary_op_kind
    end
  end

  class Lval
    attr_accessor :host, :offset
    protected :host=, :offset=

    def Lval.read
      lval = Lval.new
      lval.host = Host.read
      lval.offset = Offset.read
      lval
    end
    
    def accept(v)
      host.accept v
      offset.accept v
    end
    
    def identifier
      offset.is_a?(OfstNone) ? host.identifier : "(unknown)"
    end
  end

  class Host
    HOST_VARIABLE = 0
    HOST_OBJECT_AT_EFFECTIVE_ADDRESS = 1

    def Host.read
      case XDR::Int.read
      when HOST_VARIABLE: VarHost.read
      when HOST_OBJECT_AT_EFFECTIVE_ADDRESS: ExprHost.read
      else raise "Invalid discriminator for Host"
      end
    end
  end

  class VarHost < Host
    attr_accessor :id
    protected :id=

    def VarHost.read
      var_host = VarHost.new
      var_host.id = XDR::Int.read
      var_host
    end
    
    def accept(v)
      v.visit_host self
    end
    
    def identifier
      $funcs[id].name
    end
  end

  class ExprHost < Host
    attr_accessor :expr
    protected :expr=

    def ExprHost.read
      expr_host = ExprHost.new
      expr_host.expr = Expr.read
      expr_host
    end
    
    def accept(v)
      v.visit_host self
      expr.accept v
    end
    
    def identifier
      "(unknown)"
    end
  end

  class Offset
    OFFSET_NONE = 0
    OFFSET_FIELD_ACCESS = 1
    OFFSET_ARRAY_INDEX = 2

    def Offset.read
      case XDR::Int.read
      when OFFSET_NONE: OfstNone.read
      when OFFSET_FIELD_ACCESS: OfstField.read
      when OFFSET_ARRAY_INDEX: OfstArray.read
      else raise "Invalid discriminator for Offset"
      end
    end
  end
  
  class OfstNone < Offset
    def OfstNone.read
      OfstNone.new
    end
    
    def accept(v)
      v.visit_offset self
    end
  end
  
  class FieldAcc
    attr_accessor :field_name, :host_comp_name
    protected :field_name=, :host_comp_name=

    def FieldAcc.read
      field_acc = FieldAcc.new
      field_acc.field_name = XDR::String.read
      field_acc.host_comp_name = XDR::String.read
      field_acc
    end
  end

  class OfstField < Offset
    attr_accessor :acc, :next_ofst
    protected :acc=, :next_ofst=

    def OfstField.read
      ofst_field = OfstField.new
      ofst_field.acc = FieldAcc.read
      ofst_field.next_ofst = Offset.read
      ofst_field
    end
    
    def accept(v)
      v.visit_offset self
      next_ofst.accept v
    end
  end

  class OfstArray < Offset
    attr_accessor :index, :next_ofst
    protected :index=, :next_ofst=

    def OfstArray.read
      ofst_array = OfstArray.new
      ofst_array.index = Expr.read
      ofst_array.next_ofst = Offset.read
      ofst_array
    end
    
    def accept(v)
      v.visit_offset self
      index.accept v
      next_ofst.accept v
    end
  end

  class Init
    INIT_SINGLE = 0
    INIT_COMPOUND = 1

    def Init.read
      case XDR::Int.read
      when INIT_SINGLE: SingleInit.read
      when INIT_COMPOUND: CompInit.read
      else raise "Invalid discriminator for Init"
      end
    end
  end
  
  class SingleInit < Init
    attr_accessor :init
    protected :init=
    
    def SingleInit.read
      single_init = SingleInit.new
      single_init.init = Expr.read
      single_init
    end
    
    def accept(v)
      v.visit_init self
      init.accept v
    end
  end

  class InitWithOfst
    attr_accessor :offset, :init
    protected :offset=, :init=

    def InitWithOfst.read
      init_with_ofst = InitWithOfst.new
      init_with_ofst.offset = Offset.read
      init_with_ofst.init = Init.read
      init_with_ofst
    end
    
    def accept(v)
      offset.accept v
      init.accept v
    end
  end

  class CompInit < Init
    attr_accessor :type, :inits
    protected :type=, :inits=

    def CompInit.read
      comp_init = CompInit.new
      comp_init.type = Type.read
      comp_init.inits = XDR::Array.read InitWithOfst
      comp_init
    end
    
    def accept(v)
      v.visit_init self
      type.accept v
      inits.each { |i| i.accept v }
    end
  end

  class Label
    LABEL_REAL_LABEL = 0
    LABEL_CASE_STATEMENT = 1
    LABEL_DEFAULT_STATEMENT = 2

    def Label.read
      case XDR::Int.read
      when LABEL_REAL_LABEL: RealLbl.read
      when LABEL_CASE_STATEMENT: CaseLbl.read
      when LABEL_DEFAULT_STATEMENT: DefaultLbl.read
      else raise "Invalid discriminator for Label"
      end
    end
  end

  class RealLbl < Label
    attr_accessor :text, :loc, :in_orig_source
    protected :text=, :loc=, :in_orig_source=

    def RealLbl.read
      real_lbl = RealLbl.new
      real_lbl.text = XDR::String.read
      real_lbl.loc = Location.read
      real_lbl.in_orig_source = XDR::Bool.read
      real_lbl
    end
    
    def accept(v)
      v.visit_label self
    end
  end

  class CaseLbl < Label
    attr_accessor :expr, :loc
    protected :expr=, :loc=

    def CaseLbl.read
      case_lbl = CaseLbl.new
      case_lbl.expr = Expr.read
      case_lbl.loc = Location.read
      case_lbl
    end
    
    def accept(v)
      v.visit_label self
      expr.accept v
    end
  end
  
  class DefaultLbl < Label
    attr_accessor :loc
    protected :loc=
    
    def DefaultLbl.read
      default_lbl = DefaultLbl.new
      default_lbl.loc = Location.read
      default_lbl
    end
    
    def accept(v)
      v.visit_label self
    end
  end

  class Stmt
    attr_accessor :labels, :content, :cfg_id, :cfg_successors
    attr_accessor :cfg_predecessors
    protected :labels=, :content=, :cfg_id=, :cfg_successors=
    protected :cfg_predecessors=

    def Stmt.read
      stmt = Stmt.new
      stmt.labels = XDR::Array.read Label
      stmt.content = RealStmt.read
      stmt.cfg_id = XDR::Int.read
      stmt.cfg_successors = XDR::Array.read Stmt
      stmt.cfg_predecessors = XDR::Array.read Stmt
      stmt
    end
    
    def accept(v)
      v.visit_stmt self
      labels.each { |l| l.accept v }
      content.accept v
      cfg_successors.each { |s| s.accept v }
      cfg_predecessors.each { |p| p.accept v }
    end
  end

  class RealStmt
    STMT_INSTRUCTIONS = 0
    STMT_RETURN = 1
    STMT_GOTO = 2
    STMT_BREAK = 3
    STMT_CONTINUE = 4
    STMT_IF = 5
    STMT_SWITCH = 6
    STMT_INFINITE_LOOP = 7
    STMT_BLOCK = 8
    STMT_MSVC_TRY_FINALLY = 9
    STMT_MSVC_TRY_EXCEPT = 10

    def RealStmt.read
      case XDR::Int.read
      when STMT_INSTRUCTIONS: Instructions.read
      when STMT_RETURN: ReturnStmt.read
      when STMT_GOTO: GotoStmt.read
      when STMT_BREAK: BreakStmt.read
      when STMT_CONTINUE: ContinueStmt.read
      when STMT_IF: IfStmt.read
      when STMT_SWITCH: SwitchStmt.read
      when STMT_INFINITE_LOOP: LoopStmt.read
      when STMT_BLOCK: StmtBlock.read
      when STMT_MSVC_TRY_FINALLY: TryFinally.read
      when STMT_MSVC_TRY_EXCEPT: TryExcept.read
      else raise "Invalid discriminator for RealStmt"
      end
    end
  end

  class Instructions < RealStmt
    attr_accessor :instrs
    protected :instrs=

    def Instructions.read
      instructions = Instructions.new
      instructions.instrs = XDR::Array.read Instr
      instructions
    end
    
    def accept(v)
      v.visit_real_stmt self
      instrs.each { |i| i.accept v }
    end
  end

  class ReturnStmt < RealStmt
    attr_accessor :value, :loc
    protected :value=, :loc=

    def ReturnStmt.read
      return_stmt = ReturnStmt.new
      return_stmt.value = XDR::Option.read Expr
      return_stmt.loc = Location.read
      return_stmt
    end
    
    def accept(v)
      v.visit_real_stmt self
      value.accept v if value
    end
  end

  class GotoStmt < RealStmt
    attr_accessor :target_label, :loc
    protected :target_label=, :loc=

    def GotoStmt.read
      goto_stmt = GotoStmt.new
      goto_stmt.target_label = Label.read
      goto_stmt.loc = Location.read
      goto_stmt
    end
    
    def accept(v)
      v.visit_real_stmt self
      target_label.accept v
    end
  end

  class BreakStmt < RealStmt
    attr_accessor :loc
    protected :loc=

    def BreakStmt.read
      break_stmt = BreakStmt.new
      break_stmt.loc = Location.read
      break_stmt
    end
    
    def accept(v)
      v.visit_real_stmt self
    end
  end

  class ContinueStmt < RealStmt
    attr_accessor :loc
    protected :loc=

    def ContinueStmt.read
      continue_stmt = ContinueStmt.new
      continue_stmt.loc = Location.read
      continue_stmt
    end
    
    def accept(v)
      v.visit_real_stmt self
    end
  end

  class IfStmt < RealStmt
    attr_accessor :guard, :then_branch, :else_branch, :loc
    protected :guard=, :then_branch=, :else_branch=, :loc=

    def IfStmt.read
      if_stmt = IfStmt.new
      if_stmt.guard = Expr.read
      if_stmt.then_branch = Block.read
      if_stmt.else_branch = Block.read
      if_stmt.loc = Location.read
      if_stmt
    end
    
    def accept(v)
      v.visit_real_stmt self
      guard.accept v
      then_branch.accept v
      else_branch.accept v
    end
  end

  class SwitchStmt < RealStmt
    attr_accessor :guard, :block, :cases, :loc
    protected :guard=, :block=, :cases=, :loc=

    def SwitchStmt.read
      switch_stmt = SwitchStmt.new
      switch_stmt.guard = Expr.read
      switch_stmt.block = Block.read
      switch_stmt.cases = XDR::Array.read Stmt
      switch_stmt.loc = Location.read
      switch_stmt
    end
    
    def accept(v)
      v.visit_real_stmt self
      guard.accept v
      block.accept v
      cases.each { |c| c.accept v }
    end
  end

  class LoopStmt < RealStmt
    attr_accessor :body, :cfg_continue, :cfg_break, :loc
    protected :body=, :cfg_continue=, :cfg_break=, :loc=

    def LoopStmt.read
      loop_stmt = LoopStmt.new
      loop_stmt.body = Block.read
      loop_stmt.cfg_continue = XDR::Option.read Stmt
      loop_stmt.cfg_break = XDR::Option.read Stmt
      loop_stmt.loc = Location.read
      loop_stmt
    end
    
    def accept(v)
      v.visit_real_stmt self
      body.accept v
      cfg_continue.accept v if cfg_continue
      cfg_break.accept v if cfg_break
    end
  end

  class StmtBlock < RealStmt
    attr_accessor :block
    protected :block=

    def StmtBlock.read
      stmt_block = StmtBlock.new
      stmt_block.block = Block.read
      stmt_block
    end
    
    def accept(v)
      v.visit_real_stmt self
      block.accept v
    end
  end

  class TryFinally < RealStmt
    attr_accessor :try, :finally, :loc
    protected :try=, :finally=, :loc=

    def TryFinally.read
      try_finally = TryFinally.new
      try_finally.try = Block.read
      try_finally.finally = Block.read
      try_finally.loc = Location.read
      try_finally
    end
    
    def accept(v)
      v.visit_real_stmt self
      try.accept v
      finally.accept v
    end
  end

  class FilterSeq
    attr_accessor :discarded_values, :real_value
    protected :discarded_values=, :real_value=

    def FilterSeq.read
      filter_seq = FilterSeq.new
      filter_seq.discarded_values = XDR::Array.read Instr
      filter_seq.real_value = Expr.read
      filter_seq
    end
    
    def accept(v)
      discarded_values.each { |d| d.accept v }
      real_value.accept v
    end
  end

  class TryExcept < RealStmt
    attr_accessor :try, :filter, :except, :loc
    protected :try=, :filter=, :except=, :loc=

    def TryExcept.read
      try_except = TryExcept.new
      try_except.try = Block.read
      try_except.filter = FilterSeq.read
      try_except.except = Block.read
      try_except.loc = Location.read
      try_except
    end
    
    def accept(v)
      v.visit_real_stmt self
      try.accept v
      filter.accept v
      except.accept v
    end
  end

  class Instr
    INSTR_ASSIGNMENT = 0
    INSTR_FUNCTION_CALL = 1
    INSTR_INLINE_ASM = 2

    def Instr.read
      case XDR::Int.read
      when INSTR_ASSIGNMENT: Assignment.read
      when INSTR_FUNCTION_CALL: FunCall.read
      when INSTR_INLINE_ASM: InlineASM.read
      else raise "Invalid discriminator for Instr"
      end
    end
  end

  class Assignment < Instr
    attr_accessor :lvalue, :contents, :loc
    protected :lvalue=, :contents=, :loc=

    def Assignment.read
      assignment = Assignment.new
      assignment.lvalue = Lval.read
      assignment.contents = Expr.read
      assignment.loc = Location.read
      assignment
    end
    
    def accept(v)
      v.visit_instr self
      lvalue.accept v
      contents.accept v
    end
  end

  class FunCall < Instr
    attr_accessor :result_lvalue, :name, :args, :loc
    protected :result_lvalue=, :name=, :args=, :loc=

    def FunCall.read
      fun_call = FunCall.new
      fun_call.result_lvalue = XDR::Option.read Lval
      fun_call.name = Expr.read
      fun_call.args = XDR::Array.read Expr
      fun_call.loc = Location.read
      fun_call
    end
    
    def accept(v)
      v.visit_instr self
      result_lvalue.accept v if result_lvalue
      name.accept v
      args.each { |a| a.accept v }
    end
  end

  class ASMOut
    attr_accessor :name, :constraint, :destination
    protected :name=, :constraint=, :destination=

    def ASMOut.read
      asm_out = ASMOut.new
      asm_out.name = XDR::Option.read XDR::String
      asm_out.constraint = XDR::String.read
      asm_out.destination = Lval.read
      asm_out
    end
    
    def accept(v)
      destination.accept v
    end
  end

  class ASMIn
    attr_accessor :name, :constraint, :loaded_input
    protected :name=, :constraint=, :loaded_input=

    def ASMIn.read
      asm_in = ASMIn.new
      asm_in.name = XDR::Option.read XDR::String
      asm_in.constraint = XDR::String.read
      asm_in.loaded_input = Expr.read
      asm_in
    end
    
    def accept(v)
      loaded_input.accept v
    end
  end

  class InlineASM < Instr
    attr_accessor :volatile, :templates, :output_operands, :input_operands
    attr_accessor :clobbered_registers, :loc
    protected :volatile=, :templates=, :output_operands=, :input_operands=
    protected :clobbered_registers=, :loc=

    def InlineASM.read
      inline_asm = InlineASM.new
      inline_asm.volatile = XDR::Bool.read
      inline_asm.templates = XDR::Array.read XDR::String
      inline_asm.output_operands = XDR::Array.read ASMOut
      inline_asm.input_operands = XDR::Array.read ASMIn
      inline_asm.clobbered_registers = XDR::Array.read XDR::String
      inline_asm.loc = Location.read
      inline_asm
    end
    
    def accept(v)
      v.visit_instr self
      output_operands.each { |o| o.accept v }
      input_operands.each { |i| i.accept v }
    end
  end

  class Location
    attr_accessor :line, :file, :byte
    protected :line=, :file=, :byte=

    def Location.read
      location = Location.new
      location.line = XDR::Int.read
      location.file = XDR::String.read
      location.byte = XDR::Int.read
      location
    end
  end

  class Typesig
    attr_accessor :typesig
    protected :typesig=

    def Typesig.read
      typesig = Typesig.new
      typesig.typesig = XDR::Int.read
      typesig
    end
    
    def accept(v)
      nil
    end
  end

end
