#!/usr/bin/env ruby

# = Name
#
# query.rb -- decode and extract information from a C AST in XDR format
#
# = Synopsis
#
# query.rb <OPTIONS> <AST FILE> [STRINGS FILE]
# 
# If no strings file is given, it is assumed to have the same name as the AST
# file, except ending in .strings.xdr instead of just .xdr.
#
# = Options
#
# At least one option should be provided to indicate a query to execute, or
# the program will simply print this help. Multiple options can be given, and
# the associated queries will be executed in that order.
#
# -h, --help::
#   Show this help.
#
# -c fn, --func-calls-in fn::
#   Print each function call that occurs in the function fn, along with the
#   location of the call.
#
# -C fn, --calls-to-func fn::
#   Print each call of the function fn that occurs in any function, along 
#   with the location of the call.
#
# -g fn, --global-uses-in fn::
#   Print each use of any global variable that occurs in fn, along with the
#   location of the use.
#
# -G var, --uses-of-global var::
#   Print each use of the global variable var that occurs in any function,
#   along with the location of the use.

require "getoptlong" # to parse command-line options
require "rdoc/usage" # to print the help information given above

require "xdr" # to read the basic XDR primitive types
require "ast" # to read and work with the C AST

# This Visitor implements every visit method that could possibly be called by
# the AST, but it does nothing. It has two purposes: to document all the
# visit methods and to make other Visitors more concise. For example, if your
# Visitor only cares to look at function definitions, you can simply inherit
# from this class and only override the visit_global method.
class NilVisitor
  def visit_file(file); end
  def visit_global(global); end
  def visit_local_var_decl(var); end
  def visit_type(type); end
  def visit_attribute(attrib); end
  def visit_attr_param(param); end
  def visit_expr(expr); end
  def visit_constant(const); end
  def visit_host(host); end
  def visit_offset(offset); end
  def visit_init(init); end
  def visit_label(label); end
  def visit_stmt(stmt); end
  def visit_real_stmt(real_stmt); end
  def visit_instr(instr); end
end

# This Visitor handles the --func-calls-in (-c) query. If you construct one,
# give it a function name parameter, and pass it to the accept method of an
# AST, then it will find all the function calls that occur inside the
# function you specified. You'll need to invoke print_results to see what was
# found, or you'll just get a header.
class FuncCallsInVisitor < NilVisitor
  def initialize(func_name)
    @func_name = func_name
    @func_calls = []
    @in_desired_func = false
    header = "Functions called in #{@func_name}:"
    border = "-" * header.length
    puts border, header, border
  end

  def visit_global(global)
    if global.is_a?(AST::FuncDefn) && global.name == @func_name
      @in_desired_func = true
    else
      @in_desired_func = false
    end
  end

  def visit_instr(instr)
    if @in_desired_func && instr.is_a?(AST::FunCall)
      name = instr.name.identifier
      file = instr.loc.file
      line = instr.loc.line
      @func_calls << "#{name} @ #{file}:#{line}"
    end
  end
  
  def print_results
    @func_calls.uniq!
    if @func_calls.empty?
      puts "None"
    else
      @func_calls.each { |call| puts call }
    end
  end
end

# This Visitor handles the --calls-to-func (-c) query. If you construct one,
# give it a function name parameter, and pass it to the accept method of an
# AST, then it will find all the calls to the function you specified that
# occur anywhere. You'll need to invoke print_results to see what was found,
# or you'll just get a header.
class CallsToFuncVisitor < NilVisitor
  def initialize(func_name)
    @func_name = func_name
    @func_calls = []
    @current_func = ""
    header = "Functions that call #{@func_name}:"
    border = "-" * header.length
    puts border, header, border
  end
  
  def visit_global(global)
    if global.is_a?(AST::FuncDefn)
      @current_func = global.name
    else
      @current_func = ""
    end
  end
  
  def visit_instr(instr)
    if instr.is_a?(AST::FunCall) && instr.name.identifier == @func_name
      file = instr.loc.file
      line = instr.loc.line
      @func_calls << "#{@current_func} @ #{file}:#{line}"
    end
  end
  
  def print_results
    @func_calls.uniq!
    if @func_calls.empty?
      puts "None"
    else
      @func_calls.each { |call| puts call }
    end
  end
end

def read_strings_file(strings)
  $strings = {}
  $in = File.open strings, "r"
  len = XDR::Int.read
  len.times do
    str = XDR::String.really_read
    key = XDR::Int.read
    $strings[key] = str
  end
  $in.close
end

def read_ast_file(ast)
  $funcs = []
  $gvars = []
  $in = File.open ast, "r"
  AST::File.read
end

def print_func_calls_in(ast, func_name)
  visitor = FuncCallsInVisitor.new func_name
  ast.accept visitor
  visitor.print_results
end

def print_calls_to_func(ast, func_name)
  visitor = CallsToFuncVisitor.new func_name
  ast.accept visitor
  visitor.print_results
end

opts = GetoptLong.new(
  ["--help", "-h", GetoptLong::NO_ARGUMENT],
  ["--func-calls-in", "-c", GetoptLong::REQUIRED_ARGUMENT],
  ["--calls-to-func", "-C", GetoptLong::REQUIRED_ARGUMENT],
  ["--global-uses-in", "-g", GetoptLong::REQUIRED_ARGUMENT],
  ["--uses-of-global", "-G", GetoptLong::REQUIRED_ARGUMENT]
)

queries = []
opts.each do |opt, arg|
  case opt
  when "--help": RDoc::usage
  when "--func-calls-in": queries << [:print_func_calls_in, arg]
  when "--calls-to-func": queries << [:print_calls_to_func, arg]
  when "--global-uses-in": queries << [:print_global_uses_in, arg]
  when "--uses-of-global": queries << [:print_uses_of_global, arg]
  end
end
RDoc::usage if queries.empty?

ast_filename = nil
strings_filename = nil
case ARGV.length
when 1
  ast_filename = ARGV.shift
  strings_filename = ast_filename.sub(/\.xdr$/, ".strings.xdr")
when 2
  ast_filename = ARGV.shift
  strings_filename = ARGV.shift
else RDoc::usage
end

read_strings_file strings_filename
ast = read_ast_file ast_filename

queries.each do |q|
  query = q.shift
  arg = q.shift
  Object.method(query).call(ast, arg)
end
