# Readers for the basic XDR primitive types.

module XDR

  class Int
    # String.unpack can only read unsigned ints (in guaranteed big-endian
    # form, anyway), so convert those to signed when regular ints are used.
    def Int.read
      bit_len = 32
      max = 2 ** bit_len - 1
      mid = 2 ** (bit_len - 1)
      unsigned = UInt.read
      unsigned >= mid ? -((unsigned ^ max) + 1) : unsigned
    end
  end

  class UInt
    def UInt.read
      $in.sysread(4).unpack("N").first
    end
  end

  class Bool
    def Bool.read
      case Int.read
      when 0: false
      when 1: true
      else raise "Invalid discriminator for Bool"
      end
    end
  end

  # A 64-bit int.
  class Hyper
    def Hyper.read
      $in.sysread(8).unpack("Q").first
    end
  end

  class Double
    def Double.read
      $in.sysread(8).unpack("G").first
    end
  end

  class String
    # "Read" a string by instead reading an int ID and looking that up in a
    # table of interned strings.
    def String.read
      id = UInt.read
      str = $strings[id]
      str
    end

    # Actually read a string (used when building the table for the above).
    def String.really_read
      len = UInt.read
      pad_len = 4 - len % 4
      total_len = pad_len < 4 ? len + pad_len : len
      $in.sysread(total_len).slice(0, len)
    end
  end

  class Array
    def Array.read(cls)
      arr = []
      len = UInt.read
      len.times { arr << cls.send(:read) }
      arr
    end
  end

  class Option
    NONE = 0
    SOME = 1

    def Option.read(cls)
      case Int.read
      when NONE: nil
      when SOME: cls.send :read
      else raise "Invalid discriminator for Option"
      end
    end
  end

end
