open Rtypes

let oc = ref stdout

let strings = Hashtbl.create 1024
let next_string_id = ref 0

(****************************************************************************
 * WRITERS FOR THE NECESSARY XDR PRIMITIVE TYPES                            *
 ***************************************************************************)

let write_int i =
  output_string !oc (int4_as_string (int4_of_int i))

let write_bool b =
  write_int (if b then 1 else 0)

let write_hyper i64 =
  output_string !oc (int8_as_string (int8_of_int64 i64))

let write_double f =
  output_string !oc (fp8_as_string (fp8_of_float f))

let write_string s =
  try
    let id = Hashtbl.find strings s in
    write_int id
  with Not_found -> begin
    let id = !next_string_id in
    write_int id;
    Hashtbl.add strings s id;
    incr next_string_id
  end

let really_write_string s =
  let len = String.length s in
  output_string !oc (uint4_as_string (uint4_of_int len));
  let pad_len = 4 - len mod 4 in
  let pad = if pad_len < 4 then String.make pad_len '\000' else "" in
  output_string !oc (s ^ pad)

let write_array writer lst =
  write_int (List.length lst);
  List.iter writer lst

let write_option writer = function
  | Some content ->
      write_int 1;
      writer content
  | None ->
      write_int 0

(****************************************************************************
 * WRITERS FOR THE C AST TYPES (AS DEFINED IN C_AST.X)                      *
 ***************************************************************************)

let rec write_file f =
  write_string f.Cil.fileName;
  write_array write_global f.Cil.globals

and write_global = function
  | Cil.GType (ti, loc) ->
      write_int 0; (* GLOBAL_TYPE_DEFN *)
      write_string ti.Cil.tname;
      write_bool ti.Cil.treferenced;
      write_c_type ti.Cil.ttype;
      write_location loc
  | Cil.GCompTag (ci, loc) ->
      write_int 1; (* GLOBAL_COMPOSITE_TAG_DEFN *)
      write_string ci.Cil.cname;
      write_int (if ci.Cil.cstruct then 0 else 1);
      write_int ci.Cil.ckey;
      write_bool ci.Cil.creferenced;
      let write_field_decl fi =
        write_string fi.Cil.fname;
        write_option write_int fi.Cil.fbitfield;
        write_c_type fi.Cil.ftype;
        write_array write_attribute fi.Cil.fattr;
        write_location loc
      in
      write_array write_field_decl ci.Cil.cfields;
      write_array write_attribute ci.Cil.cattr;
      write_location loc
  | Cil.GCompTagDecl (ci, loc) ->
      write_int 2; (* GLOBAL_COMPOSITE_TAG_DECL *)
      write_string ci.Cil.cname;
      write_int (if ci.Cil.cstruct then 0 else 1);
      write_int ci.Cil.ckey;
      write_bool ci.Cil.cdefined;
      write_bool ci.Cil.creferenced;
      write_array write_attribute ci.Cil.cattr;
      write_location loc
  | Cil.GEnumTag (ei, loc) ->
      write_int 3; (* GLOBAL_ENUM_TAG_DEFN *)
      write_string ei.Cil.ename;
      write_bool ei.Cil.ereferenced;
      let write_enum_item (s, e, loc) =
        write_string s;
        write_expr e;
        write_location loc
      in
      write_array write_enum_item ei.Cil.eitems;
      write_array write_attribute ei.Cil.eattr;
      write_location loc
  | Cil.GEnumTagDecl (ei, loc) ->
      write_int 4; (* GLOBAL_ENUM_TAG_DECL *)
      write_string ei.Cil.ename;
      write_bool ei.Cil.ereferenced;
      write_array write_attribute ei.Cil.eattr;
      write_location loc
  | Cil.GVarDecl (vi, loc) ->
      (* A GVarDecl can represent a variable declaration or a function
       * prototype. Since the XDR distinguishes these, check the type to see
       * which one to write. *)
      begin match vi.Cil.vtype with
        | Cil.TFun (_, _, _, _) ->
            write_int 8; (* GLOBAL_FUNCTION_DECL *)
            write_string vi.Cil.vname;
            write_int vi.Cil.vid;
            write_bool vi.Cil.vinline;
            write_bool vi.Cil.vaddrof;
            write_bool vi.Cil.vreferenced;
            write_c_type vi.Cil.vtype;
            write_storage_class vi.Cil.vstorage;
            write_array write_attribute vi.Cil.vattr;
            write_location vi.Cil.vdecl
        | _ ->
            write_int 6; (* GLOBAL_VARIABLE_DECL *)
            write_var_decl vi
      end
  | Cil.GVar (vi, ii, loc) ->
      write_int 5; (* GLOBAL_VARIABLE_DEFN *)
      write_string vi.Cil.vname;
      write_int vi.Cil.vid;
      write_bool vi.Cil.vaddrof;
      write_bool vi.Cil.vreferenced;
      write_c_type vi.Cil.vtype;
      write_storage_class vi.Cil.vstorage;
      write_array write_attribute vi.Cil.vattr;
      write_location loc;
      write_option write_init ii.Cil.init
  | Cil.GFun (fd, loc) ->
      write_int 7; (* GLOBAL_FUNCTION_DEFN *)
      write_string fd.Cil.svar.Cil.vname;
      write_int fd.Cil.svar.Cil.vid;
      write_bool fd.Cil.svar.Cil.vinline;
      write_bool fd.Cil.svar.Cil.vaddrof;
      write_bool fd.Cil.svar.Cil.vreferenced;
      write_c_type fd.Cil.svar.Cil.vtype;
      write_storage_class fd.Cil.svar.Cil.vstorage;
      write_array write_var_decl fd.Cil.sformals;
      write_array write_var_decl fd.Cil.slocals;
      write_block fd.Cil.sbody;
      write_array write_attribute fd.Cil.svar.Cil.vattr;
      write_location loc
  | Cil.GAsm (tmpl, loc) ->
      write_int 9; (* GLOBAL_ASM_STATEMENT *)
      write_string tmpl;
      write_location loc
  | Cil.GPragma (attr, loc) ->
      write_int 10; (* GLOBAL_PRAGMA *)
      write_attribute attr;
      write_location loc
  | Cil.GText s ->
      write_int 11; (* GLOBAL_TEXT *)
      write_string s

and write_storage_class = function
  | Cil.NoStorage -> write_int 0
  | Cil.Static -> write_int 1
  | Cil.Register -> write_int 2
  | Cil.Extern -> write_int 3

and write_var_decl vi =
  write_string vi.Cil.vname;
  write_int vi.Cil.vid;
  write_bool vi.Cil.vglob;
  write_bool vi.Cil.vaddrof;
  write_bool vi.Cil.vreferenced;
  write_c_type vi.Cil.vtype;
  write_storage_class vi.Cil.vstorage;
  write_array write_attribute vi.Cil.vattr;
  write_location vi.Cil.vdecl

and write_block b =
  write_array write_attribute b.Cil.battrs;
  write_array write_stmt b.Cil.bstmts

and write_int_kind = function
  | Cil.IChar -> write_int 0
  | Cil.ISChar -> write_int 1
  | Cil.IUChar -> write_int 2
  | Cil.IInt -> write_int 3
  | Cil.IUInt -> write_int 4
  | Cil.IShort -> write_int 5
  | Cil.IUShort -> write_int 6
  | Cil.ILong -> write_int 7
  | Cil.IULong -> write_int 8
  | Cil.ILongLong -> write_int 9
  | Cil.IULongLong -> write_int 10

and write_float_kind = function
  | Cil.FFloat -> write_int 0
  | Cil.FDouble -> write_int 1
  | Cil.FLongDouble -> write_int 2

and write_c_type = function
  | Cil.TVoid attrs ->
      write_int 0; (* TYPE_VOID *)
      write_array write_attribute attrs
  | Cil.TInt (ik, attrs) ->
      write_int 1; (* TYPE_INT *)
      write_int_kind ik;
      write_array write_attribute attrs
  | Cil.TFloat (fk, attrs) ->
      write_int 2; (* TYPE_FLOAT *)
      write_float_kind fk;
      write_array write_attribute attrs
  | Cil.TPtr (t, attrs) ->
      write_int 3; (* TYPE_POINTER *)
      write_c_type t;
      write_array write_attribute attrs
  | Cil.TArray (t, len_opt, attrs) ->
      write_int 4; (* TYPE_ARRAY *)
      write_c_type t;
      write_option write_expr len_opt;
      write_array write_attribute attrs
  | Cil.TFun (result, formals, varargs, attrs) ->
      write_int 5; (* TYPE_FUNCTION *)
      write_bool varargs;
      write_c_type result;
      let write_parameter (_, t, attrs) =
        write_c_type t;
        write_array write_attribute attrs
      in 
      begin match formals with
        | Some params ->
            begin match params with
              | [] ->
                  write_int 0 (* NO_PARAMETERS *)
              | _ ->
                  write_int 1; (* SOME_PARAMETERS *)
                  write_array write_parameter params
            end
        | None ->
            write_int 2 (* PARAMETERS_UNKNOWN *)
      end;
      write_array write_attribute attrs
  | Cil.TNamed (ti, attrs) ->
      write_int 6; (* TYPE_NAMED *)
      write_string ti.Cil.tname;
      write_array write_attribute attrs
  | Cil.TComp (ci, attrs) ->
      write_int 7; (* TYPE_COMPOSITE *)
      write_string ci.Cil.cname;
      write_array write_attribute attrs
  | Cil.TEnum (ei, attrs) ->
      write_int 8; (* TYPE_ENUMERAETED *)
      write_string ei.Cil.ename;
      write_array write_attribute attrs
  | Cil.TBuiltin_va_list attrs ->
      write_int 9; (* TYPE_GCC_BUILTIN_VA_LIST *)
      write_array write_attribute attrs

and write_attribute = function
  | Cil.Attr (name, params) ->
      write_string name;
      write_array write_attr_param params

and write_attr_param = function
  | Cil.AInt i ->
      write_int 0; (* PARAM_INT *)
      write_int i
  | Cil.AStr s ->
      write_int 1; (* PARAM_STRING *)
      write_string s
  | Cil.ACons (name, params) ->
      write_int 2; (* PARAM_CONSTRUCTED *)
      write_string name;
      write_array write_attr_param params
  | Cil.ASizeOf t ->
      write_int 3; (* PARAM_SIZEOF_TYPE *)
      write_c_type t
  | Cil.ASizeOfE param ->
      write_int 4; (* PARAM_SIZEOF_PARAM *)
      write_attr_param param
  | Cil.ASizeOfS ts ->
      (* This is only used for attributes inside typsigs, and those don't
       * appear in the tree; they're generated by functions we don't use. *)
      assert false
  | Cil.AAlignOf t ->
      write_int 5; (* PARAM_ALIGNOF_TYPE *)
      write_c_type t
  | Cil.AAlignOfE param ->
      write_int 6; (* PARAM_ALIGNOF_PARAM *)
      write_attr_param param
  | Cil.AAlignOfS ts ->
      assert false (* Again, no typsigs. *)
  | Cil.AUnOp (op, param) ->
      write_int 7; (* PARAM_UNARY_OP *)
      write_unary_op_kind op;
      write_attr_param param
  | Cil.ABinOp (op, param1, param2) ->
      write_int 8; (* PARAM_BINARY_OP *)
      write_binary_op_kind op;
      write_attr_param param1;
      write_attr_param param2
  | Cil.ADot (param, name) ->
      write_int 9; (* PARAM_DOT *)
      write_attr_param param;
      write_string name
  | Cil.AStar param ->
      write_int 10; (* PARAM_STAR *)
      write_attr_param param
  | Cil.AAddrOf param ->
      write_int 11; (* PARAM_ADDROF *)
      write_attr_param param
  | Cil.AIndex (base, offset) ->
      write_int 12; (* PARAM_INDEX *)
      write_attr_param base;
      write_attr_param offset
  | Cil.AQuestion (guard, tbranch, fbranch) ->
      write_int 13; (* PARAM_TERNARY_OP *)
      write_attr_param guard;
      write_attr_param tbranch;
      write_attr_param fbranch

and write_expr = function
  | Cil.Const c ->
      write_int 0; (* EXPR_CONST *)
      write_constant c
  | Cil.Lval lv ->
      write_int 1; (* EXPR_LVALUE *)
      write_lval lv
  | Cil.SizeOf t ->
      write_int 2; (* EXPR_SIZEOF_TYPE *)
      write_c_type t
  | Cil.SizeOfE e ->
      write_int 3; (* EXPR_SIZEOF_EXPR *)
      write_expr e
  | Cil.SizeOfStr s ->
      write_int 4; (* EXPR_SIZEOF_STRING_LITERAL *)
      write_string s
  | Cil.AlignOf t ->
      write_int 5; (* EXPR_ALIGNOF_TYPE *)
      write_c_type t
  | Cil.AlignOfE e ->
      write_int 6; (* EXPR_ALIGNOF_EXPR *)
      write_expr e
  | Cil.UnOp (op, e, t) ->
      write_int 7; (* EXPR_UNARY_OP *)
      write_unary_op_kind op;
      write_expr e;
      write_c_type t
  | Cil.BinOp (op, e1, e2, t) ->
      write_int 8; (* EXPR_BINARY_OP *)
      write_binary_op_kind op;
      write_expr e1;
      write_expr e2;
      write_c_type t
  | Cil.CastE (t, e) ->
      write_int 9; (* EXPR_TYPE_CAST *)
      write_c_type t;
      write_expr e
  | Cil.AddrOf lv ->
      write_int 10; (* EXPR_ADDROF *)
      write_lval lv
  | Cil.StartOf lv ->
      write_int 11; (* EXPR_STARTOF *)
      write_lval lv

and write_constant = function
  | Cil.CInt64 (i64, ik, text_rep_opt) ->
      write_int 0; (* CONST_INTEGER *)
      write_hyper i64;
      write_int_kind ik;
      write_option write_string text_rep_opt;
  | Cil.CStr s ->
      write_int 1; (* CONST_STRING *)
      write_string s
  | Cil.CWStr i64s ->
      write_int 2; (* CONST_WIDE_STRING *)
      write_array write_hyper i64s
  | Cil.CChr c ->
      write_int 3; (* CONST_CHAR *)
      write_int (int_of_char c)
  | Cil.CReal (f, fk, text_rep_opt) ->
      write_int 4; (* CONST_FLOATING_POINT *)
      write_double f;
      write_float_kind fk;
      write_option write_string text_rep_opt;
  | Cil.CEnum (value, name, ei) ->
      write_int 5; (* CONST_ENUMERATION *)
      write_string name;
      write_expr value;
      write_string ei.Cil.ename

and write_unary_op_kind = function
  | Cil.Neg -> write_int 0
  | Cil.BNot -> write_int 1
  | Cil.LNot -> write_int 2

and write_binary_op_kind = function
  | Cil.PlusA -> write_int 0
  | Cil.PlusPI -> write_int 1
  | Cil.IndexPI -> write_int 2
  | Cil.MinusA -> write_int 3
  | Cil.MinusPI -> write_int 4
  | Cil.MinusPP -> write_int 5
  | Cil.Mult -> write_int 6
  | Cil.Div -> write_int 7
  | Cil.Mod -> write_int 8
  | Cil.Shiftlt -> write_int 9
  | Cil.Shiftrt -> write_int 10
  | Cil.Lt -> write_int 11
  | Cil.Gt -> write_int 12
  | Cil.Le -> write_int 13
  | Cil.Ge -> write_int 14
  | Cil.Eq -> write_int 15
  | Cil.Ne -> write_int 16
  | Cil.BAnd -> write_int 17
  | Cil.BXor -> write_int 18
  | Cil.BOr -> write_int 19
  | Cil.LAnd -> write_int 20
  | Cil.LOr -> write_int 21

and write_lval (host, offset) =
  begin match host with
    | Cil.Var vi ->
        write_int 0; (* HOST_VARIABLE *)
        write_int vi.Cil.vid
    | Cil.Mem e ->
        write_int 1; (* HOST_OBJECT_AT_EFFECTIVE_ADDRESS *)
        write_expr e
  end;
  write_offset offset

and write_offset = function
  | Cil.NoOffset ->
      write_int 0; (* OFFSET_NONE *)
  | Cil.Field (fi, offset) ->
      write_int 1; (* OFFSET_FIELD_ACCESS *)
      write_string fi.Cil.fname;
      write_string fi.Cil.fcomp.Cil.cname;
      write_offset offset
  | Cil.Index (e, offset) ->
      write_int 2; (* OFFSET_ARRAY_INDEX *)
      write_expr e;
      write_offset offset

and write_init = function
  | Cil.SingleInit e ->
      write_int 0; (* INIT_SINGLE *)
      write_expr e
  | Cil.CompoundInit (t, inits) ->
      write_int 1; (* INIT_COMPOUND *)
      write_c_type t;
      write_array (fun (o, i) -> write_offset o; write_init i) inits

and write_label = function
  | Cil.Label (s, loc, in_orig_src) ->
      write_int 0; (* LABEL_REAL_LABEL *)
      write_string s;
      write_location loc;
      write_bool in_orig_src
  | Cil.Case (e, loc) ->
      write_int 1; (* LABEL_CASE_STATEMENT *)
      write_expr e;
      write_location loc
  | Cil.Default loc ->
      write_int 2; (* LABEL_DEFAULT_STATEMENT *)
      write_location loc

and write_stmt stmt =
  write_array write_label stmt.Cil.labels;
  write_real_stmt stmt.Cil.skind;
  write_int stmt.Cil.sid;
  write_array write_stmt stmt.Cil.succs;
  write_array write_stmt stmt.Cil.preds

and write_real_stmt = function
  | Cil.Instr instrs ->
      write_int 0; (* STMT_INSTRUCTIONS *)
      write_array write_instr instrs
  | Cil.Return (value_opt, loc) ->
      write_int 1; (* STMT_RETURN *)
      write_option write_expr value_opt;
      write_location loc
  | Cil.Goto (s_ref, loc) ->
      write_int 2; (* STMT_GOTO *)
      (* We only really need to write one real label to uniquely identify
       * !s_ref, so just use the first one in its labels list. CIL guarantees
       * that there will be at least one. *)
      let is_real_label = function
        | Cil.Label (_, _, _) -> true
        | _ -> false
      in
      write_label (List.find is_real_label !s_ref.Cil.labels);
      write_location loc
  | Cil.Break loc ->
      write_int 3; (* STMT_BREAK *)
      write_location loc
  | Cil.Continue loc ->
      write_int 4; (* STMT_CONTINUE *)
      write_location loc
  | Cil.If (guard, then_branch, else_branch, loc) ->
      write_int 5; (* STMT_IF *)
      write_expr guard;
      write_block then_branch;
      write_block else_branch;
      write_location loc
  | Cil.Switch (guard, block, cases, loc) ->
      write_int 6; (* STMT_SWITCH *)
      write_expr guard;
      write_block block;
      write_array write_stmt cases;
      write_location loc
  | Cil.Loop (body, loc, continue_opt, break_opt) ->
      write_int 7; (* STMT_INFINITE_LOOP *)
      write_block body;
      write_option write_stmt continue_opt;
      write_option write_stmt break_opt;
      write_location loc
  | Cil.Block block ->
      write_int 8; (* STMT_BLOCK *)
      write_block block
  | Cil.TryFinally (try_block, finally_block, loc) ->
      write_int 9; (* STMT_MSVC_TRY_FINALLY *)
      write_block try_block;
      write_block finally_block;
      write_location loc
  | Cil.TryExcept (try_block, (discards, value), except_block, loc) ->
      write_int 10; (* STMT_MSVC_TRY_EXCEPT *)
      write_block try_block;
      write_array write_instr discards;
      write_expr value;
      write_block except_block;
      write_location loc

and write_instr = function
  | Cil.Set (lvalue, contents, loc) ->
      write_int 0; (* INSTR_ASSIGNMENT *)
      write_lval lvalue;
      write_expr contents;
      write_location loc
  | Cil.Call (maybe_result_lvalue, name, args, loc) ->
      write_int 1; (* INSTR_FUNCTION_CALL *)
      write_option write_lval maybe_result_lvalue;
      write_expr name;
      write_array write_expr args;
      write_location loc
  | Cil.Asm (attrs, tmpls, outs, ins, clobs, loc) ->
      write_int 2; (* INSTR_INLINE_ASM *)
      begin match attrs with
        | Cil.Attr ("volatile", []) :: [] -> write_bool true
        | [] -> write_bool false
        | _ -> assert false
      end;
      write_array write_string tmpls;
      let write_out (maybe_name, constrt, dest) =
        write_option write_string maybe_name;
        write_string constrt;
        write_lval dest
      in
      write_array write_out outs;
      let write_in (maybe_name, constrt, loaded_input) =
        write_option write_string maybe_name;
        write_string constrt;
        write_expr loaded_input
      in
      write_array write_in ins;
      write_array write_string clobs;
      write_location loc

and write_location loc =
  write_int loc.Cil.line;
  write_string loc.Cil.file;
  write_int loc.Cil.byte

(****************************************************************************
 * WRITER FOR THE TABLE OF INTERNED STRINGS (AS DEFINED IN STRING_TABLE.X)  *
 ***************************************************************************)

let write_string_table () =
  write_int (Hashtbl.length strings);
  let write_table_entry s id =
    really_write_string s;
    write_int id
  in
  Hashtbl.iter write_table_entry strings

(****************************************************************************
 * PUBLIC INTERFACE                                                         *
 ***************************************************************************)

(* This is the entrance point; pass it a CIL tree and you'll get XDR. *)
let write_xdr file =
  let cil_name = file.Cil.fileName in
  let no_extn_name = String.sub cil_name 0 (String.length cil_name - 2) in
  let ast_name = no_extn_name ^ ".xdr" in
  oc := open_out_bin ast_name;
  write_file file;
  flush !oc;
  let strings_name = no_extn_name ^ ".strings.xdr" in
  oc := open_out_bin strings_name;
  write_string_table ();
  flush !oc

(* This serves to tell CIL that the entrance point is write_xdr so it knows
 * what to do when cilly is passed the command-line option --dowritexdr. *)
let feature : Cil.featureDescr = {
  Cil.fd_name = "writexdr";
  Cil.fd_enabled = ref false;
  Cil.fd_description = "serializing the AST to XDR format";
  Cil.fd_extraopt = [];
  Cil.fd_doit = write_xdr;
  Cil.fd_post_check = false;
}
