open Cil
open Escape
open Pretty
open Xmlescape
open Xmlpretty

class xmlCilPrinterClass : cilPrinter = object (self)
  (* we don't use these *)
  method setCurrentFormals _ = ()
  method setPrintInstrTerminator _ = ()
  method getPrintInstrTerminator () = ""
  
  (* cil conflates var decls and func protos, this handles both *)
  method pVDecl () vi =
    let xml_tag, xml_attrs = match vi.vtype with
      | TFun _ ->
          "funp",
          [
            ("n", vi.vname);
            ("id", string_of_int vi.vid);
            ("glob", string_of_bool vi.vglob);
            ("inln", string_of_bool vi.vinline);
            ("addr", string_of_bool vi.vaddrof);
            ("used", string_of_bool vi.vreferenced);
          ]
      | _ ->
          "vdec",
          [
            ("n", vi.vname);
            ("id", string_of_int vi.vid);
            ("glob", string_of_bool vi.vglob);
            ("addr", string_of_bool vi.vaddrof);
            ("used", string_of_bool vi.vreferenced);
          ] in
    xml_complex xml_tag xml_attrs [
      self#pType None () vi.vtype;
      self#pAttrs () vi.vattr;
      self#pStor vi.vstorage;
      self#pLoc vi.vdecl;
    ]
  
  method pVar vi =
    let xml_attrs = [
      ("n", vi.vname);
      ("id", string_of_int vi.vid);
      ("glob", string_of_bool vi.vglob);
      ("addr", string_of_bool vi.vaddrof);
      ("used", string_of_bool vi.vreferenced);
    ] in
    xml_complex "vuse" xml_attrs []
  
  method pLval () lv =
    let b, o = lv in
    let base = match b with
      | Var vi ->
          xml_complex "base" [("k", "var")] [self#pVar vi]
      | Mem e ->
          xml_complex "base" [("k", "obj")] [
            self#pExp () e;
          ] in
    xml_complex "lv" [] [
      base;
      self#pOffset nil o;
    ]
  
  (* ignore base arg, i don't use it but cilPrinter interface wants it *)
  method pOffset _ = function
    | NoOffset -> nil
    | Field (fi, o) ->
        xml_complex "ofst" [("k", "fld")] [
          self#pFieldAcc fi;
          self#pOffset nil o;
        ]
    | Index (e, o) ->
        xml_complex "ofst" [("k", "aray")] [
          self#pExp () e;
          self#pOffset nil o;
        ]
  
  method pInstr () = function
    | Set (lv, e, loc) ->
        xml_complex "i" [("k", "assn")] [
          self#pLval () lv;
          self#pExp () e;
          self#pLoc loc;
        ]
    | Call (lv_opt, e, args, loc) -> begin match lv_opt with
        (* if call has a result lval it's really an assignment *)
        | Some lv ->
            (* might need to cast result to type of lvalue. code to check for
             * this case adapted from defaultCilPrinterClass#pInstr *)
            let dest_t = typeOfLval lv in
            let pt = typeSig in
            let call_maybe_casted = match unrollType (typeOf e) with
              | TFun (ret_t, _, _, _)
                when not (Util.equals (pt ret_t) (pt dest_t)) ->
                  xml_complex "cast" [] [
                    self#pType None () dest_t;
                    self#pFunCall e args;
                  ]
              | _ ->
                  self#pFunCall e args in
            xml_complex "i" [("k", "assn")] [
              self#pLval () lv;
              call_maybe_casted;
              self#pLoc loc;
            ]
        | None ->
          xml_complex "i" [("k", "func")] [
            self#pFunCall e args;
            self#pLoc loc;
          ]
        end (* match lv_opt *)
    (* escape everything, god knows what's in what of these *)
    | Asm (attrs, tmpls, outs, ins, clobs, loc) ->
        let pOut (name_opt, constrnt, lv) =
          let maybe_name = match name_opt with
            | Some name -> xml_simple "n" [] [text (xml_escape_string name)]
            | None -> nil in
          xml_complex "op" [] [
            maybe_name;
            xml_simple "cs" [] [
              text (xml_escape_string constrnt);
            ];
            xml_complex "od" [] [self#pLval () lv];
          ] in
        let pIn (name_opt, constrnt, e) =
          let maybe_name = match name_opt with
            | Some name -> xml_simple "n" [] [text (xml_escape_string name)]
            | None -> nil in
          xml_complex "op" [] [
            maybe_name;
            xml_simple "cs" [] [
              text (xml_escape_string constrnt);
            ];
            xml_complex "li" [] [self#pExp () e];
          ] in
        let safe_tmpls = List.map xml_escape_string tmpls in
        let safe_clobs = List.map xml_escape_string clobs in
        xml_complex "i" [("k", "asm")] [
          self#pAttrs () attrs;
          xml_complex "tmps" [] [
            xml_seq (xml_simple "tmpl" []) text safe_tmpls;
          ];
          xml_complex "oops" [] [seq nil pOut outs];
          xml_complex "iops" [] [seq nil pIn ins];
          xml_complex "regs" [] [
            xml_seq (xml_simple "reg" []) text safe_clobs;
          ];
          self#pLoc loc;
        ]
  
  (* cil labels include case/default in switch statements *)
  method pLabel () = function
    | Label (name, loc, non_cil) ->
        let xml_attrs = [
          ("n", name);
          ("cil", string_of_bool (not non_cil));
        ] in
        xml_complex "lbl" xml_attrs [
          self#pLoc loc;
        ]
    | Case (e, loc) ->
        xml_complex "case" [] [
          xml_complex "grd" [] [self#pExp () e];
          self#pLoc loc;
        ]
    | Default loc ->
        xml_complex "dflt" [] [self#pLoc loc]
  
  method pStmt () s =
    let maybe_succs =
      if s.succs = [] then nil
      else xml_complex "cfgs" [] [seq nil (self#pStmt ()) s.succs] in
    let maybe_preds =
      if s.preds = [] then nil
      else
        xml_complex "cfgp" [] [seq nil (self#pStmt ()) s.preds] in
    let xml_attrs =
      if s.sid = -1 then []
      else [("cfid", string_of_int s.sid)] in
    xml_complex "s" xml_attrs [
      seq nil (self#pLabel ()) s.labels;
      (* actual stmt content is in here. first param is unused *)
      self#pStmtKind invalidStmt () s.skind;
      maybe_preds;
      maybe_succs;
    ]
  
  method dStmt out ind s =
    fprint out !lineLength (indent ind (self#pStmt () s))
  
  method dBlock out ind b =
    fprint out !lineLength (indent ind (self#pBlock () b))
  
  method pBlock () b =
    xml_complex "b" [] [
      self#pAttrs () b.battrs;
      seq nil (self#pStmt ()) b.bstmts;
    ]
  
  method pGlobal () = function
    | GType (ti, loc) ->
        let xml_attrs = [
          ("n", ti.tname);
          ("used", string_of_bool ti.treferenced);
        ] in
        xml_complex "type" xml_attrs [ 
          self#pType None () ti.ttype;
          self#pLoc loc;
        ]
    | GCompTag (ci, loc) ->
        let xml_attrs = [
          ("n", ci.cname);
          ("k", if ci.cstruct then "stru" else "unio");
          ("key", string_of_int ci.ckey);
          ("defn", string_of_bool ci.cdefined);
          ("used", string_of_bool ci.creferenced);
        ] in
        xml_complex "cdef" xml_attrs [
          seq nil (self#pFieldDecl ()) ci.cfields;
          self#pAttrs () ci.cattr;
          self#pLoc loc;
        ]
    | GCompTagDecl (ci, loc) ->
        let xml_attrs = [
          ("n", ci.cname);
          ("k", if ci.cstruct then "stru" else "unio");
          ("key", string_of_int ci.ckey);
          ("defn", string_of_bool ci.cdefined);
          ("used", string_of_bool ci.creferenced);
        ] in
        xml_complex "cdec" xml_attrs [
          self#pAttrs () ci.cattr;
          self#pLoc loc;
        ]
    | GEnumTag (ei, loc) ->
        let xml_attrs = [
          ("n", ei.ename);
          ("used", string_of_bool ei.ereferenced);
        ] in
        xml_complex "edef" xml_attrs [
          seq nil self#pEnumItem ei.eitems;
          self#pAttrs () ei.eattr;
          self#pLoc loc;
        ]
    | GEnumTagDecl (ei, loc) ->
        let xml_attrs = [
          ("n", ei.ename);
          ("used", string_of_bool ei.ereferenced);
        ] in
        xml_complex "edec" xml_attrs [
          self#pAttrs () ei.eattr;
          self#pLoc loc;
        ]
    (* loc part is ignored because it's already in vi.vdecl... right? *)
    | GVarDecl (vi, _) -> self#pVDecl () vi
    | GVar (vi, ii, loc) ->
        let init = match ii.init with
          | Some init -> self#pInit () init
          | None -> nil in
        let xml_attrs = [
          ("n", vi.vname);
          ("id", string_of_int vi.vid);
          ("addr", string_of_bool vi.vaddrof);
          ("used", string_of_bool vi.vreferenced);
        ] in
        xml_complex "vdef" xml_attrs [
          self#pType None () vi.vtype;
          self#pAttrs () vi.vattr;
          self#pStor vi.vstorage;
          init;
          self#pLoc loc;
        ]
    | GFun (fd, loc) ->
        let pFunVar xml_tag vi =
          let xml_attrs = [
            ("n", vi.vname);
            ("id", string_of_int vi.vid);
            ("addr", string_of_bool vi.vaddrof);
            ("used", string_of_bool vi.vreferenced);
          ] in
          xml_complex xml_tag xml_attrs [
            self#pType None () vi.vtype;
            self#pAttrs () vi.vattr;
            self#pStor vi.vstorage;
            self#pLoc vi.vdecl;
          ] in
        let maybe_cfg_info =
          if fd.sallstmts = [] then nil
          else 
            xml_complex "cfg" [] [
              seq nil (self#pStmt ()) fd.sallstmts;
            ] in
        let xml_attrs = [
          ("n", fd.svar.vname);
          ("id", string_of_int fd.svar.vid);
          ("inln", string_of_bool fd.svar.vinline);
          ("addr", string_of_bool fd.svar.vaddrof);
          ("used", string_of_bool fd.svar.vreferenced);
        ] in
        xml_complex "fdef" xml_attrs [
          self#pType None () fd.svar.vtype;
          self#pAttrs () fd.svar.vattr;
          self#pStor fd.svar.vstorage;
          xml_complex "args" [] [
            seq nil (pFunVar "arg") fd.sformals;
          ];
          xml_complex "vars" [] [
            seq nil (pFunVar "loc") fd.slocals;
          ];
          xml_complex "body" [] [self#pBlock () fd.sbody];
          maybe_cfg_info;
          self#pLoc loc;
        ]
    | GAsm (tmpl, loc) ->
        xml_complex "gasm" [] [
          (* could be nasty chars in here, escape to be safe *)
          xml_simple "tmpl" [] [text (xml_escape_string tmpl)];
          self#pLoc loc;
        ]
    | GPragma (attr, loc) ->
        xml_complex "pgma" [] [
          (* should this really use pAttr? *)
          fst (self#pAttr attr);
          self#pLoc loc;
        ]
    | GText str ->
        xml_simple "text" [] [text str]
  
  method dGlobal out g =
    fprint out !lineLength (indent 2 (self#pGlobal () g))
  
  method pFieldDecl () fi =
    let bf, maybe_width = match fi.fbitfield with
      | Some width -> true, xml_simple "widt" [] [num width]
      | None -> false, nil in
    let xml_attrs = [
      ("n", fi.fname);
      ("bitf", string_of_bool bf);
    ] in
    xml_complex "fldd" xml_attrs [
      self#pType None () fi.ftype;
      maybe_width;
      self#pAttrs () fi.fattr;
      self#pLoc fi.floc;
    ]
  
  (* first arg not needed in this class *)
  method pType _ () = function
    | TVoid attrs ->
        xml_complex "t" [("k", "void")] [self#pAttrs () attrs]
    | TInt (ik, attrs) ->
        let xml_attrs = [("k", self#string_of_ikind ik)] in
        xml_complex "t" xml_attrs [self#pAttrs () attrs]
    | TFloat (fk, attrs) ->
        let xml_attrs = [("k", self#string_of_fkind fk)] in
        xml_complex "t" xml_attrs [self#pAttrs () attrs]
    | TPtr (t, attrs) ->
        xml_complex "t" [("k", "ptr")] [
          self#pType None () t;
          self#pAttrs () attrs;
        ]
    | TArray (t, e_opt, attrs) ->
        let maybe_length = match e_opt with
          | Some e -> xml_complex "len" [] [self#pExp () e]
          | None -> nil in
        xml_complex "t" [("k", "aray")] [
          self#pType None () t;
          maybe_length;
          self#pAttrs () attrs;
        ]
    | TFun (result, formals_opt, varargs, attrs) ->
        let pFormal (name, t, attrs) =
          xml_complex "arg" [] [
            self#pType None () t;
            self#pAttrs () attrs;
          ] in
        let maybe_formals = match formals_opt with
          | Some formals ->
              if formals = [] then
                (* foo(void) *)
                xml_simple "none" [] []
              else
                (* foo(some params) *)
                seq nil pFormal formals
          | None ->
              (* proto of foo not seen yet *)
              xml_simple "unk" [] [] in
        let xml_attrs = [
          ("k", "fun");
          ("varg", string_of_bool varargs);
        ] in
        xml_complex "t" xml_attrs [
          xml_complex "res" [] [self#pType None () result];
          maybe_formals;
          self#pAttrs () attrs;
        ]
    | TNamed (ti, attrs) ->
        xml_complex "t" [("k", "namd"); ("n", ti.tname)] [
          (* self#pType None () ti.ttype; *)
          self#pAttrs () attrs;
        ]
    | TComp (ci, attrs) ->
        let xml_attrs = [
          ("n", ci.cname);
          ("k", if ci.cstruct then "stru" else "unio");
          ("key", string_of_int ci.ckey);
          ("defn", string_of_bool ci.cdefined);
        ] in
        xml_complex "t" xml_attrs [
          self#pAttrs () attrs;
        ]
    | TEnum (ei, attrs) ->
        let xml_attrs = [
          ("n", ei.ename);
          ("k", "enum");
        ] in
        xml_complex "t" xml_attrs [
          self#pAttrs () attrs;
        ]
    (* wtf is this? 15 google results for it *)
    | TBuiltin_va_list attrs ->
        xml_complex "t" [("k", "gccv")] [
          self#pAttrs () attrs;
        ]
  
  method pAttr = function
    | Attr (name, params) ->
        xml_complex "a" [("n", name)] [
          seq nil (self#pAttrParam ()) params;
        ], true (* we don't use this boolean but interface wants it *)
  
  method pAttrParam () = function
    | AInt i ->
        xml_simple "prm" [("k", "int")] [num i]
    | AStr s ->
        xml_simple "prm" [("k", "str")] [text s]
    | ACons (name, params) ->
        xml_complex "prm" [("k", "cons"); ("n", name)] [
          seq nil (self#pAttrParam ()) params;
        ]
    | ASizeOf t ->
        xml_complex "prm" [("k", "sz")] [self#pType None () t]
    | ASizeOfE param ->
        xml_complex "prm" [("k", "sz")] [
          self#pAttrParam () param;
        ]
    | ASizeOfS ts ->
        xml_complex "prm" [("k", "sz")] [self#pTypeSig ts]
    | AAlignOf t ->
        xml_complex "prm" [("k", "algn")] [self#pType None () t]
    | AAlignOfE param ->
        xml_complex "prm" [("k", "algn")] [
          self#pAttrParam () param;
        ]
    | AAlignOfS ts ->
        xml_complex "prm" [("k", "algn")] [self#pTypeSig ts]
    | AUnOp (uop, param) ->
        let xml_attrs = [
          ("k", "unop");
          ("op", self#string_of_unop uop);
        ] in
        xml_complex "prm" xml_attrs [self#pAttrParam () param]
    | ABinOp (bop, param1, param2) ->
        let xml_attrs = [
          ("k", "bnop");
          ("op", self#string_of_binop bop);
        ] in
        xml_complex "prm" xml_attrs [
          self#pAttrParam () param1;
          self#pAttrParam () param2;
        ]
    (* what is this dot? a field access? is star a ptr deref? *)
    | ADot (param, name) ->
        xml_complex "prm" [("k", "dot"); ("fldn", name)] [
          self#pAttrParam () param;
        ]
    | AStar param ->
        xml_complex "prm" [("k", "star")] [self#pAttrParam () param]
    | AAddrOf param ->
        xml_complex "prm" [("k", "addr")] [
          self#pAttrParam () param;
        ]
    | AIndex (ar, ind) ->
        xml_complex "prm" [("k", "aray")] [
          xml_complex "aray" [] [self#pAttrParam () ar];
          xml_complex "ind" [] [self#pAttrParam () ind];
        ]
    | AQuestion (g, tb, fb) ->
        xml_complex "prm" [("k", "tnop")] [
          xml_complex "grd" [] [self#pAttrParam () g];
          xml_complex "tbch" [] [self#pAttrParam () tb];
          xml_complex "fbch" [] [self#pAttrParam () fb];
        ]
  
  method pAttrs () attrs =
    let pAttr = (fun x -> fst (self#pAttr x)) in
    if attrs = [] then nil (* don't print <attributes/> every damn time *)
    else seq nil pAttr attrs
  
  (* no # directives in xml, we use pLoc to print location info instead *)
  method pLineDirective ?forcefile _ = nil
  
  (* the actual content of a stmt is in here. stmtkind is a pretty bad name *)
  method pStmtKind _ () = function
    | Instr instrs ->
        seq nil (self#pInstr ()) instrs;
    | Return (e_opt, loc) ->
        let maybe_e = match e_opt with
          | Some e -> self#pExp () e
          | None -> nil in
        xml_complex "ret" [] [
          maybe_e;
          self#pLoc loc;
        ]
    | Goto (s_ref, loc) ->
        (* can't call self#pStmt on !s_ref here, could cause infinite loop.
         * so instead just print its labels and stmt id *)
        let s = !s_ref in
        let xml_attrs = [("id", string_of_int s.sid)] in
        let target = xml_complex "lbls" xml_attrs [
          seq nil (self#pLabel ()) s.labels;
        ] in
        xml_complex "goto" [] [
          target;
          self#pLoc loc;
        ]
    | Break loc ->
        xml_complex "brk" [] [self#pLoc loc]
    | Continue loc ->
        xml_complex "cont" [] [self#pLoc loc]
    | If (guard, tb, eb, loc) ->
        xml_complex "if" [] [
          xml_complex "grd" [] [self#pExp () guard];
          xml_complex "tbch" [] [self#pBlock () tb];
          xml_complex "ebch" [] [self#pBlock () eb];
          self#pLoc loc;
        ]
    (* info in ignored stmt list param is redundant, stmts are in b *)
    | Switch (e, b, _, loc) ->
        xml_complex "swch" [] [
          xml_complex "grd" [] [self#pExp () e];
          self#pBlock () b;
          self#pLoc loc;
        ]
    (* all loops in cil are converted to while (1) with continue/break *)
    | Loop (b, loc, cont_opt, brk_opt) ->
        (* continue and break stmts are in b somewhere, but if prepareCFG was
         * called, we have easy references to them here too *)
        let maybe_cont = match cont_opt with
          | Some cont ->
              xml_complex "cfgc" [] [
                self#pStmt () cont;
              ]
          | None -> nil in
        let maybe_brk = match brk_opt with
          | Some brk ->
              xml_complex "cfgb" [] [
                self#pStmt () brk;
              ]
          | None -> nil in
        xml_complex "loop" [] [
          self#pBlock () b;
          maybe_cont;
          maybe_brk;
          self#pLoc loc;
        ]
    | Block b -> self#pBlock () b
    | TryFinally (tb, fb, loc) ->
        xml_complex "mstf" [] [
          xml_complex "try" [] [self#pBlock () tb];
          xml_complex "fnly" [] [self#pBlock () fb];
          self#pLoc loc;
        ]
    | TryExcept (tb, (instrs, e), eb, loc) ->
        (* pretty weird stuff here. cil stores some instrs for the filter along
         * with the expected guard exp. these instrs are separated by the comma
         * operator, so they are side-effect instrs whose values are ignored *)
        let filter = xml_complex "fltr" [] [
          xml_complex "disc" [] [seq nil (self#pInstr ()) instrs];
          self#pExp () e;
        ] in
        xml_complex "mste" [] [
          xml_complex "try" [] [self#pBlock () tb];
          xml_complex "expt" [] [
            filter;
            self#pBlock () eb;
          ];
          self#pLoc loc;
        ]
  
  method pExp () = function
    | Const c ->
        xml_complex "e" [("k", "cnst")] [self#pConst c]
    | Lval lv -> 
        xml_complex "e" [("k", "lv")] [self#pLval () lv]
    | SizeOf t -> 
        xml_complex "e" [("k", "sz")] [self#pType None () t]
    | SizeOfE e ->
        xml_complex "e" [("k", "sz")] [self#pExp () e]
    | SizeOfStr s ->
        xml_complex "e" [("k", "sz")] [
          xml_simple "str" [] [text (xml_escape_string s)];
        ]
    | AlignOf t ->
        xml_complex "e" [("k", "algn")] [self#pType None () t]
    | AlignOfE e ->
        xml_complex "e" [("k", "algn")] [self#pExp () e]
    | UnOp (uop, e, t) ->
        let xml_attrs = [
          ("k", "unop");
          ("op", self#string_of_unop uop);
        ] in
        xml_complex "e" xml_attrs [
          self#pExp () e;
          self#pType None () t;
        ]
    | BinOp (bop, e1, e2, t) ->
        let xml_attrs = [
          ("k", "bnop");
          ("op", self#string_of_binop bop);
        ] in
        xml_complex "e" xml_attrs [
          self#pExp () e1;
          self#pExp () e2;
          self#pType None () t;
        ]
    | CastE (t, e) ->
        xml_complex "e" [("k", "cast")] [
          self#pType None () t;
          self#pExp () e;
        ]
    | AddrOf lv ->
        xml_complex "e" [("k", "addr")] [self#pLval () lv]
    | StartOf lv ->
        xml_complex "e" [("k", "sarr")] [self#pLval () lv]
  
  method pInit () = function
    | SingleInit e ->
        xml_complex "init" [] [self#pExp () e]
    | CompoundInit (t, inits_with_kind) ->
        let pOffInit = function
          | Index(e, NoOffset), i ->
              xml_complex "indx" [] [
                self#pExp () e;
                self#pInit () i;
              ]
          | Field(fi, NoOffset), i ->
              xml_complex "fld" [] [
                self#pFieldAcc fi; (* right function to use here? *)
                self#pInit () i;
              ]
          (* cil docs say the above are the only valid cases *)
          | _ -> assert false in
        xml_complex "cmpi" [] [
          self#pType None () t;
          seq nil pOffInit inits_with_kind;
        ]
  
  method dInit out ind i =
    fprint out !lineLength (indent ind (self#pInit () i))
  
  (* end cilPrinter interface methods, begin private helper methods *)
  
  method private pStor s =
    let cls = match s with
      | NoStorage -> "def"
      | Static -> "stic"
      | Register -> "rgst"
      | Extern -> "extr" in
    xml_simple "scls" [] [text cls]
  
  method private pLoc l =
    let xml_attrs = [
      ("f", l.file);
      ("l", string_of_int l.line);
      ("b", string_of_int l.byte);
    ] in
    xml_simple "loc" xml_attrs []
  
  method private pFieldAcc fi =
    let bf, maybe_width = match fi.fbitfield with
      | Some width -> true, xml_simple "widt" [] [num width]
      | None -> false, nil in
    let xml_attrs = [
      ("n", fi.fname);
      ("comp", fi.fcomp.cname);
      ("bitf", string_of_bool bf);
    ] in
    xml_complex "flda" xml_attrs [maybe_width]
  
  method private pFunCall f args =
    xml_complex "func" [] [
      xml_complex "n" [] [self#pExp () f];
      xml_seq (xml_complex "arg" []) (self#pExp ()) args;
    ]
  
  method private pEnumItem (name, e, loc) =
    xml_complex "itm" [] [
      xml_simple "n" [] [text name];
      xml_complex "v" [] [self#pExp () e];
      self#pLoc loc;
    ]
  
  method private string_of_ikind = function
    | IChar -> "chr"
    | ISChar -> "schr"
    | IUChar -> "uchr"
    | IInt -> "int"
    | IUInt -> "uint"
    | IShort -> "sht"
    | IUShort -> "usht"
    | ILong -> "lng"
    | IULong -> "ulng"
    | ILongLong -> "llg"
    | IULongLong -> "ullg"
  
  method private string_of_fkind = function
    | FFloat -> "flt"
    | FDouble -> "dbl"
    | FLongDouble -> "ldbl"
  
  method private string_of_unop = function
    | Neg -> "neg"
    | BNot -> "bnot"
    | LNot -> "lnot"

  method private string_of_binop = function
    | PlusA -> "plus"
    | PlusPI -> "ppi"
    (* as cil docs say this is just a PlusPI for our purposes *)
    | IndexPI -> "ppi"
    | MinusA -> "mnus"
    | MinusPI -> "pmi"
    | MinusPP -> "pmp"
    | Mult -> "mult"
    | Div -> "div"
    | Mod -> "mod"
    | Shiftlt -> "shl"
    | Shiftrt -> "shr"
    | Lt -> "lt"
    | Gt -> "gt"
    | Le -> "le"
    | Ge -> "ge"
    | Eq -> "eq"
    | Ne -> "ne"
    | BAnd -> "band"
    | BXor -> "bxor"
    | BOr -> "bor"
    | LAnd -> "land"
    | LOr -> "lor"
  
  (* not sure what the deal is with these. defaultCilPrinterClass just prints
   * literally the text "sizeof(<typsig>)" (yes, with angle brackets), which
   * i'm pretty sure is not even valid c. so for now i will do something
   * similar, until i figure out what should actually be done here *)
  method private pTypeSig ts = xml_simple "tsig" [] []
  
  method private pConst = function
    | CInt64 (i64, ik, rep_opt) ->
        let maybe_text_rep = match rep_opt with
          | Some rep -> xml_simple "text" [] [text rep]
          | None -> nil in
        let xml_attrs = [
          ("k", self#string_of_ikind ik);
        ] in
        xml_complex "cnst" xml_attrs [
          xml_simple "v" [] [
            text (xml_escape_string (Int64.to_string i64));
          ];
          maybe_text_rep;
        ]
    | CStr s ->
        xml_complex "cnst" [("k", "str")] [
          xml_simple "v" [] [text (xml_escape_string s)];
        ]
    | CWStr i64s ->
        xml_complex "cnst" [("k", "wstr")] [
          xml_simple "v" [] [text (xml_escape_wstring i64s)];
        ]
    | CChr c ->
        xml_complex "cnst" [("k", "chr")] [
          xml_simple "v" [] [text (xml_escape_char c)];
        ]
    | CReal (f, fk, rep_opt) ->
        let maybe_text_rep = match rep_opt with
          | Some rep -> xml_simple "text" [] [text rep]
          | None -> nil in
        let xml_attrs = [
          ("k", self#string_of_fkind fk);
        ] in
        xml_complex "cnst" xml_attrs [
          xml_simple "v" [] [real f];
          maybe_text_rep;
        ]
    | CEnum (e, name, ei) ->
        let xml_attrs = [
          ("n", name);
          ("k", "enum");
          ("enum", ei.ename);
        ] in
        xml_complex "cnst" xml_attrs [
          xml_complex "v" [] [self#pExp () e];
        ]
  
end (* class xmlCilPrinterClass *)

let format_filename s =
  let r = Str.regexp "/" in
  Str.global_replace r "_" s

let write_file f =
  let printer = new xmlCilPrinterClass in
  printDepth := max_int;
  lineLength := max_int;
  fastMode := true;
  let globals_doc = List.map (fun g -> printer#pGlobal () g) f.Cil.globals in
  let filename = format_filename f.Cil.fileName in
  let cil_name = f.Cil.fileName in
  let no_extn_name = String.sub cil_name 0 (String.length cil_name - 2) in
  let ast_name = no_extn_name ^ ".xml" in
  let file_doc = xml_complex "file" ["name", filename] globals_doc in
  let xml = sprint !lineLength file_doc in
  let f = open_out_bin ast_name in
  output_string f xml;
  flush f;
  close_out f

let feature : featureDescr = {
  fd_name = "writexdr";
  fd_enabled = ref false;
  fd_description = "serializing the AST to XML format (compact)";
  fd_extraopt = [];
  fd_doit = write_file;
  fd_post_check = false;
}
